from sqlalchemy import Column, Unicode

from clld.db.meta import Base, DBSession

__all__ = ('Config',)


class Config(Base):

    """Model class to allow storage of key-value pairs of configuration data.

    This model is also (ab-)used to implement a mechanism linking database
    objects of all types without enforcing referential intagrity, e.g. to model chains
    of superseding objects, where referred objects may become obsolete themselves.
    """

    key = Column(Unicode)
    value = Column(Unicode)

    gone = '__gone__'

    @staticmethod
    def replacement_key(model, id_):
        """Determine the replacement key for an object.

        :param model: Model class or instance.
        :param id_: Identifier of a class instance.
        :return: ``str`` representation identifying a database object.
        """
        if isinstance(model, str):
            name = model
        elif isinstance(model, type):
            name = model.__name__
        else:
            name = model.__class__.__name__
        return '__%s_%s__' % (name, id_)

    @classmethod
    def get_replacement_id(cls, model, id_):
        """Lookup and retrieve the ID of an object.

        :return: id of a resource registered as replacement for the specified resource.
        """
        res = DBSession.query(cls.value)\
            .filter(cls.key == cls.replacement_key(model, id_)).first()
        if res:
            return res[0]

    @classmethod
    def add_replacement(cls, replaced, replacement, model=None, session=None):
        """Method to register a replacement relation.

        :param replaced: db object or identifier of the object to be replaced.
        :param replacement: db object or identifier of the superseding object.
        :param model: If only an identifier is passed as ``replaced`` or ``replacement``\
        the corresponding model class must be passed as ``model``.
        :param session: Db session the relation is added to.
        """
        session = session or DBSession
        value = getattr(replacement, 'id', replacement) if replacement else cls.gone
        session.add(cls(
            key=cls.replacement_key(model or replaced, getattr(replaced, 'id', replaced)),
            value=value))



from setuptools import setup, find_packages


setup(
    name='{{cookiecutter.directory_name}}',
    version='0.0',
    description='{{cookiecutter.directory_name}}',
    classifiers=[
        "Programming Language :: Python",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ],
    author='',
    author_email='',
    url='',
    keywords='web pyramid pylons',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'clld',
{% if cookiecutter.mpg == true %}
        'clldmpg',
{% endif %}
],
extras_require={
        'dev': ['flake8', 'waitress'],
        'test': [
            'mock',
            'pytest>=5.4',
            'pytest-clld',
            'pytest-mock',
            'pytest-cov',
            'coverage>=4.2',
            'selenium',
            'zope.component>=3.11.0',
        ],
    },
    test_suite="{{cookiecutter.directory_name}}",
    entry_points="""\
    [paste.app_factory]
    main = {{cookiecutter.directory_name}}:main
""")



import pytest

from clld.db.models import common
from clld.web.datatables.sentence import Sentences

from testutils import handle_dt


@pytest.mark.parametrize(
    "params",
    [
        {},
        {
            'sSearch_5': 'x',
            'sSearch_2': ' ',
            'iSortingCols': '1',
            'iSortCol_0': '5',
            'sSortDir_0': 'desc'},
        {
            'sSearch_4': 'text',
            'iSortingCols': '1',
            'iSortCol_0': '4',
            'sSortDir_0': 'desc'},
        {'parameter': 'parameter'},
        {'language': 'language'},
    ])
def test_plain(request_factory, params):
    with request_factory(params=params) as req:
        handle_dt(req, Sentences, common.Sentence)


def test_AudioCol(env, mocker):
    from clld.web.datatables.sentence import AudioCol, Sentences

    col = AudioCol(Sentences(env['request'], common.Sentence), 'audio')
    col.order()
    col.search('yes')
    col.format(mocker.Mock())



"""
Support for the provider implementation of an OLAC OAI-PMH repository.

.. seealso:: http://www.language-archives.org/OLAC/repositories.html
"""
import re
import copy
import datetime
import collections

from pyramid.renderers import render
from pyramid.response import Response
from sqlalchemy.orm import joinedload, undefer

from clld.db.models.common import Language, LanguageIdentifier, Identifier, IdentifierType
from clld.interfaces import IOlacConfig


#
# OAI-PMH protocol specifics:
#
VERBS = [
    'GetRecord',
    'Identify',
    'ListIdentifiers',
    'ListMetadataFormats',
    'ListRecords',
    'ListSets',
]

ERRORS = {
    'badArgument': 'The request includes illegal arguments, is missing required '
    'arguments, includes a repeated argument, or values for arguments have an illegal '
    'syntax.',
    'badResumptionToken': 'The value of the resumptionToken argument is invalid or '
    'expired.',
    'badVerb': 'Value of the verb argument is not a legal OAI-PMH verb, the verb '
    'argument is missing, or the verb argument is repeated.',
    'cannotDisseminateFormat': 'The metadata format identified by the value given for '
    'the metadataPrefix argument is not supported by the item or by the repository.',
    'idDoesNotExist': 'The value of the identifier argument is unknown or illegal in '
    'this repository.',
    'noRecordsMatch': 'The combination of the values of the from, until, set and '
    'metadataPrefix arguments results in an empty list.',
    'noMetadataFormats': 'There are no metadata formats available for the specified '
    'item.',
    'noSetHierarchy': 'The repository does not support sets.'
}

MD_PREFIX = 'olac'

TIMESTAMP_REGEX = r'[0-9]{4}\-[0-9]{2}\-[0-9]{2}'
TIMESTAMP_PATTERN = re.compile(TIMESTAMP_REGEX + '$')

Participant = collections.namedtuple('Participant', 'role name email')
Institution = collections.namedtuple('Institution', 'name url location')


def timestamp(dt=None):
    return str(dt or datetime.datetime.utcnow()).split('.')[0].replace(' ', 'T') + 'Z'


def date(dt=None):
    return str(dt or datetime.datetime.utcnow()).split(' ')[0]


class ResumptionToken(object):

    """Represents an OAI-PMH resumption token.

    We encode all information from a List query in the resumption token so that we do
    not actually have to keep track of sequences of requests (in the spirit of REST).

    .. seealso: http://www.openarchives.org/OAI/openarchivesprotocol.html#FlowControl
    """

    PATTERN = re.compile('(?P<offset>[0-9]+)(?P<from>f%s)?(?P<until>u%s)?$'
                         % (TIMESTAMP_REGEX, TIMESTAMP_REGEX))
    limit = 100

    def __init__(self, url_arg=None, offset=None, from_=None, until=None):
        def datetime_from_iso(s):
            return datetime.datetime(*map(int, s.split('-')))

        self.offset = offset or 0
        self.from_ = from_
        self.until = until

        if url_arg is not None:
            m = self.PATTERN.match(url_arg)
            assert m
            self.offset = int(m.group('offset'))
            assert self.offset % self.limit == 0
            if m.group('from'):
                self.from_ = datetime_from_iso(m.group('from')[1:])
            if m.group('until'):
                self.until = datetime_from_iso(m.group('until')[1:]) + datetime.timedelta(1)

    def __str__(self):
        res = "%s" % self.offset
        if self.from_:
            res += "f%s" % date(self.from_)
        if self.until:
            res += "u%s" % date(self.until)
        assert self.PATTERN.match(res)
        return res


class OlacConfig(object):

    """Configuration of an applications OLAC repository."""

    scheme = 'oai'
    delimiter = ':'

    def _query(self, req):
        subquery = req.db.query(Identifier)\
            .filter_by(type=IdentifierType.iso.value)\
            .join(LanguageIdentifier)\
            .filter_by(language_pk=Language.pk)
        return req.db.query(Language).filter(subquery.exists())\
            .options(undefer('updated'),
                     joinedload(Language.languageidentifier).
                     joinedload(LanguageIdentifier.identifier))

    def get_earliest_record(self, req):
        return self._query(req).order_by(Language.updated, Language.pk).first()

    def get_record(self, req, identifier):
        rec = Language.get(self.parse_identifier(req, identifier), default=None)
        assert rec
        return rec

    def query_records(self, req, from_=None, until=None):
        q = self._query(req).order_by(Language.pk)
        if from_:
            q = q.filter(Language.updated >= from_)
        if until:
            q = q.filter(Language.updated < until)
        return q

    def format_identifier(self, req, item):
        return self.delimiter.join([self.scheme, req.dataset.domain, item.id])

    def parse_identifier(self, req, id_):
        assert self.delimiter in id_
        return id_.split(self.delimiter)[-1]

    def admin(self, req):
        """Configure the archive participant with role admin.

        Note: According to http://www.language-archives.org/OLAC/repositories.html the
        list of participants
        > must include the system administrator whose email address is given in the
        > <oai:adminEmail> element of the Identify response.

        :param req: The current request.
        :return: A suitable `Participant` instance or None.
        """
        return Participant("Admin", "Archive Admin", req.contact_email_address)

    def description(self, req):
        # Note: According to http://www.language-archives.org/OLAC/repositories.html the
        # list of participants
        # > must include the system administrator whose email address is given in the
        # > <oai:adminEmail> element of the Identify response.
        participants = [self.admin(req)]
        for ed in req.dataset.editors:
            participants.append(Participant(
                "Editor",
                ed.contributor.name,
                ed.contributor.email or req.contact_email_address))
        return {
            'archiveURL': 'http://%s/' % req.dataset.domain,
            'participants': participants,
            'institution': Institution(
                req.dataset.publisher_name,
                req.dataset.publisher_url,
                req.dataset.publisher_place,
            ),
            'synopsis': req.dataset.description or '',
        }


def olac(req):
    """View implementing the OLAC OAI-PMH repository protocol."""
    return olac_with_cfg(req, req.registry.getUtility(IOlacConfig))


def olac_with_cfg(req, cfg):
    """Factory function for olac views with different configurations.

    If applications want to disseminate metadata for other resources than languages
    this function can be used to provide a second olac repository.
    """
    res = dict(verb=None, error=None, response_date=timestamp(), params={}, date=date)
    res['cfg'] = cfg

    def response(res):
        return Response(
            render('olac.mako', res, request=req).encode('utf8'),
            content_type='text/xml',
            charset=None)

    def error(_error):
        if _error:
            res['error'] = (_error, ERRORS[_error])
        return response(res)

    args = dict(req.params.items())
    res['params'] = copy.copy(args)
    res['verb'] = args.pop('verb', None)

    if res['verb'] not in VERBS:
        return error("badVerb")

    if res['verb'] == 'ListSets':
        return error("noSetHierarchy")

    if res['verb'] == 'Identify':
        if args:
            return error("badArgument")
        res['earliest'] = res['cfg'].get_earliest_record(req)

    if res['verb'] == 'GetRecord':
        if sorted(args.keys()) != ['identifier', 'metadataPrefix']:
            return error("badArgument")

        if args['metadataPrefix'] != MD_PREFIX:
            return error("cannotDisseminateFormat")

        try:
            res['language'] = res['cfg'].get_record(req, args['identifier'])
        except AssertionError:
            return error("idDoesNotExist")

    if res['verb'] in ['ListIdentifiers', 'ListRecords']:
        if (('metadataPrefix' not in args and 'resumptionToken' not in args)
                or ('from' in args and not TIMESTAMP_PATTERN.match(args['from']))
                or ('until' in args and not TIMESTAMP_PATTERN.match(args['until']))):
            return error("badArgument")

        if [arg for arg in args if arg not in [
                'from', 'until', 'metadataPrefix', 'set', 'resumptionToken']]:
            return error("badArgument")

        if 'set' in args:
            return error("noSetHierarchy")

        if 'resumptionToken' in args:
            if len(args) > 1:
                return error("badArgument")

            try:
                rt = ResumptionToken(url_arg=args['resumptionToken'])
            except AssertionError:
                return error("badResumptionToken")
        else:
            rt = ResumptionToken(
                None, 0, args.get('from', None), args.get('until', None))

        q = res['cfg'].query_records(req, from_=rt.from_, until=rt.until)
        res['languages'] = q.offset(rt.offset).limit(rt.limit).all()
        if not res['languages']:
            return error('noRecordsMatch')

        if len(res['languages']) < rt.limit:
            res['resumptionToken'] = None
        else:
            rt.offset += rt.limit
            res['resumptionToken'] = rt

    if res['verb'] == 'ListMetadataFormats':
        if args and 'identifier' not in args:
            return error("badArgument")

    return response(res)




def test_compute_language_sources(data, persist):
    from clld.db.util import compute_language_sources
    from clld.db.models.common import Source, Sentence, Language, SentenceReference

    s = Sentence(id='sentenced', language=Language(id='newlang'))
    persist(SentenceReference(sentence=s, source=Source.first()))
    compute_language_sources()


def test_compute_number_of_values(data):
    from clld.db.util import compute_number_of_values
    compute_number_of_values()


def test_icontains(data):
    from clld.db.util import icontains
    from clld.db.models.common import Dataset
    from clld.db.meta import DBSession

    for qs, count in [('Se', 1), ('^d$', 0), ('^d', 1), ('setä$', 1), ('\\\\b', 0)]:
        q = DBSession.query(Dataset).filter(icontains(Dataset.name, qs))
        assert q.count() == count



from clld.db.models import common

from testutils import handle_dt


def test_Contributions(env):
    from clld.web.datatables.contribution import Contributions

    handle_dt(env['request'], Contributions, common.Contribution)



"""
view callables implementing the sitemap protocol.

.. seealso:: http://www.sitemaps.org/
"""
import operator
import itertools

from pyramid.response import Response
from pyramid.httpexceptions import HTTPNotFound

from sqlalchemy import join, and_, true

from clld import RESOURCES
from clld.db.meta import DBSession
from clld.db.models import common
from clld.web.util.helpers import get_url_template


# http://www.sitemaps.org/protocol.html#index
LIMIT = 50000


def robots(req):
    """robots.txt response listing the sitemaps.

    .. seealso:: http://www.sitemaps.org/protocol.html#submit_robots
    """
    spec = ''
    deny = (req.registry.settings.get('clld.robots_deny') or '').strip().split()
    for robot in deny:
        spec = spec + 'User-agent: {}\nDisallow: /\n'.format(robot)
    return Response(
        "Sitemap: {}\n{}".format(req.route_url('sitemapindex'), spec), content_type="text/plain")


def _query(req, rsc):
    """Ordered sqlalchemy query.

    We must make sure, each query is ordered, so that limit and offset does make sense.
    """
    return DBSession.query(rsc.model.id, rsc.model.updated).order_by(rsc.model.pk)


def _e(name, *content):
    return '<{0}>{1}</{0}>'.format(name, ''.join(content))


def _response(type_, itemiter):
    def serialize(item):
        name = 'url' if type_ == 'urlset' else 'sitemap'
        return _e(name, *[_e(k, v) for k, v in item.items()])

    return Response(
        """\
<?xml version="1.0" encoding="UTF-8"?>
<{0} xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
{1}
</{0}>""".format(type_, '\n'.join(map(serialize, itemiter))),
        content_type="application/xml")


def sitemapindex(req):
    """Response listing resource-specific sitemaps.

    .. seealso:: http://www.sitemaps.org/protocol.html#index
    """
    def _iter(sitemaps):
        for r in RESOURCES:
            if r.with_index and r.name in sitemaps:
                n, m = divmod(_query(req, r).count(), LIMIT)
                if m:
                    n += 1
                for i in range(n):
                    yield dict(loc=req.route_url('sitemap', rsc=r.name, n=i))

    return _response(
        'sitemapindex', _iter(req.registry.settings.get('clld.sitemaps', [])))


def sitemap(req):
    """Resource-specific sitemap.

    .. note:: The resource is looked up using the URL parameter ``rsc``.

    .. seealso:: http://www.sitemaps.org/protocol.html#xmlTagDefinitions
    """
    def _iter():
        for r in RESOURCES:
            if r.name == req.matchdict['rsc']:
                query = _query(req, r)\
                    .offset(LIMIT * int(req.matchdict['n']))\
                    .limit(LIMIT)
                for id_, updated in query:
                    yield dict(
                        loc=req.route_url(r.name, id=id_),
                        lastmod=str(updated).split(' ')[0])
    return _response('urlset', _iter())


def resourcemap(req):
    """Resource-specific JSON response listing all resource instances."""
    rsc = req.params.get('rsc')
    if rsc == 'language':
        q = DBSession.query(
            common.Language.id,
            common.Language.name,
            common.Language.latitude,
            common.Language.longitude,
            common.Identifier.type.label('itype'),
            common.Identifier.name.label('iname')
        ).select_from(common.Language).outerjoin(join(
            common.LanguageIdentifier,
            common.Identifier, and_(
                common.LanguageIdentifier.identifier_pk == common.Identifier.pk,
                common.Identifier.type != 'name')
        )).filter(common.Language.active == true()).order_by(common.Language.id)

        def resources():
            for (id, name, lat, lon), rows in itertools.groupby(q, operator.itemgetter(0, 1, 2, 3)):
                identifiers = [
                    {'type': r.itype, 'identifier': r.iname.lower()
                     if r.itype.startswith('WALS') else r.iname}
                    for r in rows if r.iname is not None]
                yield {'id': id, 'name': name, 'latitude': lat, 'longitude': lon,
                       'identifiers': identifiers}
    elif rsc == 'parameter':
        q = DBSession.query(
            common.Parameter.id,
            common.Parameter.name
        ).order_by(common.Parameter.pk)

        def resources():
            for id, name in q:
                yield {'id': id, 'name': name}
    else:
        return HTTPNotFound()

    return {
        'properties': {
            'dataset': req.dataset.id,
            'uri_template': get_url_template(req, rsc, relative=False)},
        'resources': list(resources())}



from clld.db.models import common

from testutils import handle_dt


def test_Contributors(request_factory):
    from clld.web.datatables.contributor import Contributors

    with request_factory(params={
        'iSortingCols': '1',
        'iSortCol_0': '0',
        'sSortDir_0': 'desc'}
    ) as req:
        handle_dt(req, Contributors, common.Contributor)



from sqlalchemy import Column, String, Unicode, UniqueConstraint

from zope.interface import implementer
from nameparser import HumanName

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces

from .import (
    IdNameDescriptionMixin,
    DataMixin, FilesMixin, HasDataMixin, HasFilesMixin)

__all__ = ('Contributor',)


class Contributor_data(Base, DataMixin):
    pass


class Contributor_files(Base, FilesMixin):
    pass


@implementer(interfaces.IContributor)
class Contributor(Base,
                  PolymorphicBaseMixin,
                  IdNameDescriptionMixin,
                  HasDataMixin,
                  HasFilesMixin):

    """Creator of a contribution."""

    __table_args__ = (UniqueConstraint('name'),)

    url = Column(Unicode())
    email = Column(String)
    address = Column(Unicode)

    def last_first(self):
        if not self.name:
            return ''
        return '{0.last}, {0.first} {0.middle}'.format(HumanName(self.name)).strip()



"""Adapters to render bibliographic information in various formats."""
from clld.web.adapters.base import Representation
from clld.lib.bibtex import IDatabase, IRecord, Database
from clld.interfaces import IIndex, ISource, IDataTable


class Bibtex(Representation):

    """BibTeX.

    .. seealso:: http://en.wikipedia.org/wiki/BibTeX
    """

    extension = 'bib'
    mimetype = 'text/x-bibtex'

    def filename(self, ctx, req):
        return '%s-refs.%s' % (req.dataset.id, self.extension)

    def render_to_response(self, ctx, req):
        res = super(Bibtex, self).render_to_response(ctx, req)
        res.content_disposition = 'attachment; filename="%s"' % self.filename(ctx, req)
        return res

    def render(self, ctx, req):
        if ISource.providedBy(ctx):
            ctx = ctx.bibtex()
        elif IDataTable.providedBy(ctx):
            ctx = Database([s.bibtex() for s in ctx.get_query()])
        return ctx.format(self.extension)


def includeme(config):
    for interface in [IDatabase, IRecord, ISource]:
        config.register_adapter(Bibtex, interface, name=Bibtex.extension)
    config.register_adapter(Bibtex, ISource, IIndex, name=Bibtex.extension)



import pathlib

from webassets import Environment, Bundle

import clld


def skip(_in, out, **kw):
    """filter to skip content of assets which are fetched from CDN in production."""
    out.write('')  # pragma: no cover


_static_path = pathlib.Path(clld.__file__).parent.joinpath('web', 'static').as_posix()
environment = Environment(
    _static_path, '/clld:web/static/', manifest='json:', auto_build=False)
environment.append_path(_static_path, url='/clld:web/static/')

bundles = {
    'js': [
        'js/jquery.js',
        'js/leaflet.js',
        'js/Leaflet.fullscreen.js',
        'js/bootstrap.min.js',
        'js/jquery.dataTables.min.js',
        'js/oms.min.js',
        Bundle(
            'js/bootstrapx-clickover.js',
            'js/tree.jquery.js',
            'js/leaflet-providers.js',
            'js/leaflet-hash.js',
            'js/L.Control.Resizer.js',
            'js/clld.js',
            'project.js',
            filters='rjsmin'),
    ],
    'css': [
        'css/leaflet.css',
        'css/leaflet.fullscreen.css',
        'css/L.Control.Resizer.css',
        'css/clld.css',
        'css/jqtree.css',
        'css/hint.css',
        'css/jquery.dataTables.css',
        'css/bootstrap.min.css',
        'css/bootstrap-responsive.min.css',
        'project.css',
    ],
}

for name in bundles:
    environment.register(
        name, *bundles[name], **dict(output='{0}/packed.%(version)s.{0}'.format(name)))



"""Functionality to create downloads for the data of a clld app."""
import io
import csv
import gzip
import pathlib
import zipfile
import contextlib

from zope.interface import implementer
from pyramid.path import AssetResolver
from sqlalchemy.orm import joinedload, class_mapper
from clldutils.misc import format_size, to_binary

from clld.util import safe_overwrite
from clld.lib.rdf import FORMATS
from clld.web.adapters import get_adapter
from clld.web.adapters.md import TxtCitation
from clld.web.util.helpers import rdf_namespace_attrs
from clld.interfaces import IRepresentation, IDownload
from clld.db.meta import DBSession
from clld.db.models.common import Language, Source, LanguageIdentifier, Dataset
from clld.db.util import page_query

__all__ = ['download_dir', 'Download', 'N3Dump', 'CsvDump', 'RdfXmlDump']

README = """
{0} data download
{1}

Data of {0} is published under the following license:
{2}

It should be cited as

{3}
"""


def format_readme(req, dataset):
    return README.format(
        dataset.name,
        '=' * (len(dataset.name) + len(' data download')),
        dataset.license,
        TxtCitation(None).render(dataset, req))


def pkg_name(pkg):
    return pkg if isinstance(pkg, str) else pkg.__name__


def abspath(asset_spec):
    return pathlib.Path(AssetResolver().resolve(asset_spec).abspath())


def download_asset_spec(pkg, *comps):
    return '%s:%s' % (pkg_name(pkg), '/'.join(['static', 'download'] + list(comps)))


def download_dir(pkg):
    return abspath(download_asset_spec(pkg))


@implementer(IDownload)
class Download(object):

    """Represents a download format of a clld app's data."""

    ext = None

    def __init__(self, model, pkg, **kw):
        if self.ext is None:
            self.ext = kw['ext']
        self.rdf = self.ext in [fmt.extension for fmt in FORMATS.values()]
        self.model = model
        self.pkg = pkg_name(pkg)
        for k, v in kw.items():
            setattr(self, k, v)

    @property
    def name(self):
        return '%s.%s' % (class_mapper(self.model).class_.__name__.lower(), self.ext)

    def asset_spec(self, req):
        dataset = req.db.query(Dataset).first()
        return download_asset_spec(self.pkg, '%s-%s.%s' % (
            dataset.id, self.name, 'gz' if self.rdf else 'zip'))

    def url(self, req):
        return req.static_url(self.asset_spec(req))

    def abspath(self, req):
        return abspath(self.asset_spec(req))

    def size(self, req):
        _path = self.abspath(req)
        if _path.exists():
            return format_size(_path.stat().st_size)

    def label(self, req):
        return "%s [%s]" % (getattr(self, 'description', self.name), self.size(req))

    def create(self, req, filename=None, verbose=True, outfile=None):
        with safe_overwrite(outfile or self.abspath(req)) as tmp:
            if self.rdf:
                # we do not create archives with a readme for rdf downloads, because each
                # RDF entity points to the dataset and the void description of the dataset
                # covers all relevant metadata.
                #
                # TODO: write test for the file name things!?
                #
                with contextlib.closing(gzip.GzipFile(
                    filename=pathlib.Path(tmp.stem).stem, fileobj=tmp.open('wb'), mode='w',
                )) as fp:
                    self.before(req, fp)
                    for i, item in enumerate(page_query(self.query(req), verbose=verbose)):
                        self.dump(req, fp, item, i)
                    self.after(req, fp)
            else:
                with zipfile.ZipFile(tmp.as_posix(), 'w', zipfile.ZIP_DEFLATED) as zipf:
                    if not filename:
                        fp = self.get_stream()
                        self.before(req, fp)
                        for i, item in enumerate(page_query(self.query(req), verbose=verbose)):
                            self.dump(req, fp, item, i)
                        self.after(req, fp)
                        zipf.writestr(self.name, self.read_stream(fp))
                    else:  # pragma: no cover
                        zipf.write(str(pathlib.Path(filename)), self.name)
                    zipf.writestr(
                        'README.txt',
                        format_readme(req, req.db.query(Dataset).first()).encode('utf8'))

    def get_stream(self):
        return io.BytesIO()

    def read_stream(self, fp):
        fp.seek(0)
        return fp.read()

    def query(self, req):
        q = DBSession.query(self.model).filter(self.model.active == True)
        if self.model == Language:  # pragma: no cover
            q = q.options(
                joinedload(
                    Language.languageidentifier
                ).joinedload(
                    LanguageIdentifier.identifier
                )
            )
        if self.model == Source:
            q = q.options(joinedload(Source.languages))
        return q.order_by(self.model.pk)

    def before(self, req, fp):
        pass

    def dump(self, req, fp, item, index):
        adapter = get_adapter(IRepresentation, item, req, ext=self.ext)
        self.dump_rendered(req, fp, item, index, adapter.render(item, req))

    def dump_rendered(self, req, fp, item, index, rendered):
        if isinstance(rendered, str):
            rendered = rendered.encode('utf8')
        fp.write(rendered)

    def after(self, req, fp):
        pass


class CsvDump(Download):

    """Download of a resource type as csv."""

    ext = 'csv'

    def __init__(self, model, pkg, fields=None, **kw):
        """Initialize.

        fields can be a list of column names or a dictionary mapping model attribute
        names to csv column names.
        """
        super(CsvDump, self).__init__(model, pkg, **kw)
        self.fields = fields
        self.writer = None

    def get_stream(self):
        return io.StringIO(newline='')

    def read_stream(self, fp):
        return Download.read_stream(self, fp).encode('utf8')

    def get_fields(self, req):
        if not self.fields:
            self.fields = ['id', 'name']
        return self.fields

    def before(self, req, fp):
        self.writer = csv.writer(fp)
        self.writer.writerow(
            [f if isinstance(f, str) else f[1] for f in self.get_fields(req)])

    def row(self, req, fp, item, index):
        return [getattr(item, f if isinstance(f, str) else f[0])
                for f in self.get_fields(req)]

    def dump(self, req, fp, item, index):
        self.writer.writerow(self.row(req, fp, item, index))


class N3Dump(Download):

    """Download of a resource type as n-triples."""

    ext = 'n3'

    def dump_rendered(self, req, fp, item, index, rendered):
        header, body = rendered.split(to_binary('\n\n'), 1)
        if index == 0:
            fp.write(header)
            fp.write(to_binary('\n\n'))
        fp.write(body)


class RdfXmlDump(Download):

    """Download of a resource type as rdf-xml."""

    ext = 'rdf'

    def before(self, req, fp):
        fp.write(
            to_binary('<rdf:RDF ') + to_binary(rdf_namespace_attrs()) + to_binary('>\n'))

    def after(self, req, fp):
        fp.write(to_binary('</rdf:RDF>'))

    def dump_rendered(self, req, fp, item, index, rendered):
        body = rendered.split(to_binary('rdf:Description'))[1]
        fp.write(to_binary('<rdf:Description') + body + to_binary('rdf:Description>\n'))



import itertools
import collections

from clldutils.misc import nfilter
from clldutils.color import qualitative_colors
from clld.cliutil import Data, bibtex2source
from clld.db.meta import DBSession
from clld.db.models import common
from clld.lib import bibtex
{% if cookiecutter.cldf_module %}
from pycldf import Sources
{% endif %}

import {{cookiecutter.directory_name}}
from {{cookiecutter.directory_name}} import models


def main(args):
    data = Data()
    data.add(
        common.Dataset,
        {{cookiecutter.directory_name}}.__name__,
        id={{cookiecutter.directory_name}}.__name__,
        domain='{{cookiecutter.domain}}',
{% if cookiecutter.mpg == true %}
        publisher_name="Max Planck Institute for Evolutionary Anthropology",
        publisher_place="Leipzig",
        publisher_url="http://www.eva.mpg.de",
        license="http://creativecommons.org/licenses/by/4.0/",
        jsondata={
            'license_icon': 'cc-by.png',
            'license_name': 'Creative Commons Attribution 4.0 International License'},
{% else %}
        publisher_name = "",
        publisher_place = "",
        publisher_url = "",
        license = "http://creativecommons.org/licenses/by/4.0/",
        jsondata = {
            'license_icon': 'cc-by.png',
            'license_name': 'Creative Commons Attribution 4.0 International License'},
{% endif %}
    )

{% if cookiecutter.cldf_module %}
    contrib = data.add(
        common.Contribution,
        None,
        id='cldf',
        name=args.cldf.properties.get('dc:title'),
        description=args.cldf.properties.get('dc:bibliographicCitation'),
    )

    for lang in args.cldf.iter_rows('LanguageTable', 'id', 'glottocode', 'name', 'latitude', 'longitude'):
        data.add(
            models.Variety,
            lang['id'],
            id=lang['id'],
            name=lang['name'],
            latitude=lang['latitude'],
            longitude=lang['longitude'],
            glottocode=lang['glottocode'],
        )

    for rec in bibtex.Database.from_file(args.cldf.bibpath, lowercase=True):
        data.add(common.Source, rec.id, _obj=bibtex2source(rec))

    refs = collections.defaultdict(list)

{% if cookiecutter.cldf_module.lower() == 'wordlist' %}
    for param in args.cldf.iter_rows('ParameterTable', 'id', 'concepticonReference', 'name'):
        data.add(
            models.Concept,
            param['id'],
            id=param['id'],
            name='{} [{}]'.format(param['name'], param['id']),
        )
    for form in args.cldf.iter_rows('FormTable', 'id', 'form', 'languageReference', 'parameterReference', 'source'):
        vsid = (form['languageReference'], form['parameterReference'])
        vs = data['ValueSet'].get(vsid)
        if not vs:
            vs = data.add(
                common.ValueSet,
                vsid,
                id='-'.join(vsid),
                language=data['Variety'][form['languageReference']],
                parameter=data['Concept'][form['parameterReference']],
                contribution=contrib,
            )
        for ref in form.get('source', []):
            sid, pages = Sources.parse(ref)
            refs[(vsid, sid)].append(pages)
        data.add(
            common.Value,
            form['id'],
            id=form['id'],
            name=form['form'],
            valueset=vs,
        )
{% elif cookiecutter.cldf_module.lower() == 'structuredataset' %}
    for param in args.cldf.iter_rows('ParameterTable', 'id', 'name'):
        data.add(
            models.Feature,
            param['id'],
            id=param['id'],
            name='{} [{}]'.format(param['name'], param['id']),
    )
    for pid, codes in itertools.groupby(
        sorted(
            args.cldf.iter_rows('CodeTable', 'id', 'name', 'description', 'parameterReference'),
            key=lambda v: (v['parameterReference'], v['id'])),
        lambda v: v['parameterReference'],
    ):
        codes = list(codes)
        colors = qualitative_colors(len(codes))
        for code, color in zip(codes, colors):
            data.add(
                common.DomainElement,
                code['id'],
                id=code['id'],
                name=code['name'],
                description=code['description'],
                parameter=data['Feature'][code['parameterReference']],
                jsondata=dict(color=color),
            )
    for val in args.cldf.iter_rows(
            'ValueTable',
            'id', 'value', 'languageReference', 'parameterReference', 'codeReference', 'source'):
        if val['value'] is None:  # Missing values are ignored.
            continue
        vsid = (val['languageReference'], val['parameterReference'])
        vs = data['ValueSet'].get(vsid)
        if not vs:
            vs = data.add(
                common.ValueSet,
                vsid,
                id='-'.join(vsid),
                language=data['Variety'][val['languageReference']],
                parameter=data['Feature'][val['parameterReference']],
                contribution=contrib,
            )
        for ref in val.get('source', []):
            sid, pages = Sources.parse(ref)
            refs[(vsid, sid)].append(pages)
        data.add(
            common.Value,
            val['id'],
            id=val['id'],
            name=val['value'],
            valueset=vs,
            domainelement=data['DomainElement'][val['codeReference']],
        )
{% endif %}
    for (vsid, sid), pages in refs.items():
        DBSession.add(common.ValueSetReference(
            valueset=data['ValueSet'][vsid],
            source=data['Source'][sid],
            description='; '.join(nfilter(pages))
        ))
{% endif %}


def prime_cache(args):
    """If data needs to be denormalized for lookup, do that here.
    This procedure should be separate from the db initialization, because
    it will have to be run periodically whenever data has been updated.
    """



"""Adapters to serialize metadata of clld objects."""
import itertools

from zope.interface import implementer

from clld import interfaces
from clld.web.adapters.base import Representation
from clld.lib import bibtex
from clld.web.util import doi


class Metadata(Representation):

    """Virtual base class."""

    rel = 'describedby'

    @property
    def unapi_name(self):
        return getattr(self, 'unapi', self.extension)


class MetadataFromRec(Metadata):

    """Virtual base class deriving metadata from a bibtex record."""

    def rec(self, ctx, req):
        zdoi = req.registry.settings.get('clld.zenodo_version_doi') or \
            req.registry.settings.get('clld.zenodo_concept_doi')
        data = {}
        if zdoi:
            data.update(doi=zdoi, type='Data set', url=doi.url(zdoi), publisher='Zenodo')
        else:
            data.update(
                url=req.resource_url(ctx),
                address=req.dataset.publisher_place,
                publisher=req.dataset.publisher_name)

        if interfaces.IContribution.providedBy(ctx):
            genre = 'incollection'
            data['author'] = [
                c.name for c in
                itertools.chain(ctx.primary_contributors, ctx.secondary_contributors)]
            data['booktitle'] = req.dataset.description
            data['editor'] = [c.contributor.name for c in req.dataset.editors]
            id_ = '%s-%s' % (req.dataset.id, ctx.id)
        else:
            genre = 'book'
            data['editor'] = [c.contributor.name for c in ctx.editors]
            id_ = req.dataset.id

        return bibtex.Record(
            genre,
            id_,
            title='{}{}'.format(
                getattr(ctx, 'citation_name', str(ctx)),
                (' ({})'.format(req.registry.settings['clld.zenodo_version_tag'])
                 if req.registry.settings.get('clld.zenodo_version_doi') else ''),
            ),
            year=str(req.dataset.published.year),
            **data)


@implementer(interfaces.IRepresentation, interfaces.IMetadata)
class BibTex(MetadataFromRec):

    """Resource metadata as BibTex record."""

    name = 'BibTeX'
    __label__ = 'BibTeX'
    unapi = 'bibtex'
    extension = 'md.bib'
    mimetype = 'text/x-bibtex'

    def render(self, ctx, req):
        return str(self.rec(ctx, req))


@implementer(interfaces.IRepresentation, interfaces.IMetadata)
class TxtCitation(Metadata):

    """Resource metadata formatted as plain text citation."""

    name = "Citation"
    __label__ = 'Text'
    extension = 'md.txt'
    mimetype = 'text/plain'

    def render(self, ctx, req):
        if interfaces.IContribution.providedBy(ctx):
            self.template = 'contribution/md_txt.mako'
        else:  # if interfaces.IDataset.providedBy(ctx):
            self.template = 'dataset/md_txt.mako'
        return super(TxtCitation, self).render(ctx, req)



from sqlalchemy import Column, Integer, Unicode, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declared_attr

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin, HasSourceNotNullMixin)

__all__ = ('Sentence', 'SentenceReference')


class Sentence_data(Base, DataMixin):
    pass


class Sentence_files(Base, FilesMixin):
    pass


@implementer(interfaces.ISentence)
class Sentence(Base,
               PolymorphicBaseMixin,
               IdNameDescriptionMixin,
               HasDataMixin,
               HasFilesMixin):

    """Sentence of a language serving as example for some statement.

    .. note::

        Columns inherited from IdNameDescriptionMixin are interpreted as follows:

            name: The text of the sentence in object language
            description: A translation of the sentence

        Columns with a name prefix of 'markup_' can be used to store data containing HTML
        markup. These attributes should take precedence over their non-prefixed
        counterparts when rendering a Sentence.
    """

    analyzed = Column(Unicode)
    gloss = Column(Unicode)
    type = Column(Unicode)
    source = Column(Unicode)
    comment = Column(Unicode)
    original_script = Column(Unicode)
    xhtml = Column(Unicode)

    markup_text = Column(Unicode)
    markup_analyzed = Column(Unicode)
    markup_gloss = Column(Unicode)
    markup_comment = Column(Unicode)

    language_pk = Column(Integer, ForeignKey('language.pk'))

    @declared_attr
    def language(cls):
        return relationship(
            'Language', backref=backref('sentences', order_by=cls.language_pk))

    @property
    def audio(self):
        for f in self._files:
            if f.mime_type.split('/')[0] == 'audio':
                return f


class SentenceReference(Base, HasSourceNotNullMixin):

    """Association table."""

    __table_args__ = (
        UniqueConstraint('sentence_pk', 'source_pk', 'description'),
    )

    sentence_pk = Column(Integer, ForeignKey('sentence.pk'), nullable=False)
    sentence = relationship(Sentence, innerjoin=True, backref="references")



from clld.web.util.doi import *


def test_url(mocker):
    assert url('xyz').endswith('xyz')
    assert url(mocker.Mock(doi='xyz')).endswith('xyz')

    class Obj:
        jsondata = dict(doi='xyz')

    assert url(Obj()).endswith('xyz')


def test_link(env):
    assert '.png' in link(env['request'], 'xyz')


def test_badge():
    assert badge('xyz')
    assert badge('10.5281/zenodo.xyz')



import pytest
from pyramid.response import Response
from pyramid.httpexceptions import HTTPNotAcceptable, HTTPNotFound, HTTPGone, HTTPFound

from clld.db.models import common
from clld.interfaces import IDataTable

from testutils import XmlResponse


@pytest.mark.parametrize(
    "ctx_cls,model,req_props,content_type",
    [
        (None, common.Contributor, {}, None),
        (None, common.Sentence, {}, None),
        (None, common.Value, {}, None),
        (
            None,
            common.Contributor,
            {'is_xhr': True, 'params': {'sEcho': 'a'}},
            'application/json'),
        (
            None,
            common.Contributor,
            {'is_xhr': True, 'params': {'sEcho': 'a'}},
            'application/json'),
        (
            None,
            common.Contributor,
            {
                'is_xhr': False,
                'matched_route': 'contributors_alt',
                'matchdict': {'ext': 'atom'}},
            None),
    ])
def test_index_view(env, request_factory, ctx_cls, model, req_props, content_type):
    from clld.web.views import index_view

    ctx_cls = ctx_cls or env['registry'].getUtility(IDataTable, name='contributors')

    class X(ctx_cls):
        def row_class(self, item):
            return 'row-%s' % item.pk

    with request_factory(**req_props) as req:
        res = index_view(X(req, model), req)
        assert isinstance(res, Response)
        if content_type:
            assert res.content_type == content_type


def test_resource_view(env, request_factory):
    from clld.web.views import resource_view

    ctx = common.Language(id='a', name='Name')
    res = resource_view(ctx, env['request'])
    assert isinstance(res, Response)

    with request_factory(matchdict={'ext': 'x'}) as req:
        with pytest.raises(HTTPNotAcceptable):
            resource_view(ctx, req)


def test_select_combination(env, request_factory):
    from clld.web.views import select_combination

    with pytest.raises(HTTPNotFound):
        select_combination(None, env['request'])

    with request_factory(params={'parameters': 'parameter'}) as req:
        with pytest.raises(HTTPFound):
            select_combination(None, req)

    with request_factory(
        params=[('parameters', 'parameter'), ('parameters', 'no-domain')]
    ) as req:
        with pytest.raises(HTTPFound):
            select_combination(None, req)


def test__raise(env):
    from clld.web.views import _raise

    with pytest.raises(ValueError):
        _raise(None)


def test__ping(env):
    from clld.web.views import _ping

    assert _ping(None)['status'] == 'ok'


def test_js(env):
    from clld.web.views import js

    js(env['request'])


def test_gone(env):
    from clld.web.views import gone

    with pytest.raises(HTTPGone):
        gone(None, None)


def test_redirect(env):
    from clld.web.views import redirect

    with pytest.raises(HTTPFound):
        redirect(HTTPFound, lambda r: 'x', None, None)


@pytest.mark.parametrize(
    "params,assertion",
    [
        ({}, lambda res: XmlResponse(res).root.tag == 'formats'),
        ({'format': 'bib'}, lambda res: isinstance(res, HTTPNotFound)),
        ({'id': '/contributions/ccc'}, lambda res: isinstance(res, HTTPNotFound)),
        (
            {'id': '/contributions/contribution'},
            lambda res: XmlResponse(res).findall('format')),
        ({'format': 'bib', 'id': '/languages/language'}, lambda res: res),
        ({'format': 'bibtex', 'id': '/contributions/contribution'}, lambda res: res),
    ])
def test_unapi(env, request_factory, params, assertion):
    from clld.web.views import unapi

    with request_factory(params=params) as req:
        assert assertion(unapi(req))



from clld.interfaces import IIndex, IRepresentation
from clld.db.models.common import Contribution, Language, Dataset


def test_BibTex(env, env_factory):
    from clld.web.adapters import BibTex

    adapter = BibTex(None)
    assert '@' in adapter.render(Contribution.first(), env['request'])

    res = adapter.render(Dataset.first(), env_factory(zenodo_concept_doi='xyz')['request'])
    assert 'xyz' in res

    env_ = env_factory(zenodo_version_doi='xyz', zenodo_version_tag='v1.0')
    res = adapter.render(Dataset.first(), env_['request'])
    assert 'xyz' in res and 'v1.0' in res


def test_TxtCitation(env):
    from clld.web.adapters import TxtCitation

    adapter = TxtCitation(None)
    assert '.' in adapter.render(Contribution.first(), env['request'])
    assert adapter.label
    adapter.render(Dataset.first(), env['request'])


def test_Json(env):
    from clld.web.adapters.base import Json

    adapter = Json(None)
    adapter.render({'hello': 'world'}, env['request'])


def test_get_adapter(env):
    from clld.web.adapters import get_adapter

    assert get_adapter(IIndex, Language, env['request'], name='text/html') is None


def test_adapter_factory(env):
    from clld.web.adapters.base import adapter_factory

    assert IRepresentation.implementedBy(adapter_factory('template.mako'))



# flake8: noqa
from clld.web.datatables.language import Languages
from clld.web.datatables.value import Values
from clld.web.datatables.valueset import Valuesets
from clld.web.datatables.parameter import Parameters
from clld.web.datatables.unitparameter import Unitparameters
from clld.web.datatables.contribution import Contributions
from clld.web.datatables.contributor import Contributors
from clld.web.datatables.unit import Units
from clld.web.datatables.source import Sources
from clld.web.datatables.sentence import Sentences
from clld.web.datatables.unitvalue import Unitvalues



"""View logic for clld default views."""
import re
import json
import functools

from pyramid.response import Response
import pyramid.httpexceptions
from pyramid.interfaces import IRoutesMapper
from pyramid.renderers import render, render_to_response

from clld.interfaces import IRepresentation, IIndex, IMetadata
from clld.web.adapters import get_adapter, get_adapters
from clld.web.util.multiselect import MultiSelect
from clld.db.models.common import Combination


def xpartial(func, *args, **kw):
    """Augment partial to make it possible to register partials as view callables.

    >>> f = xpartial(lambda x, y: x + y, 1)
    >>> assert f(1) == 2
    """
    res = functools.partial(func, *args, **kw)
    res.__module__ = func.__module__
    return res


def redirect(cls, location, ctx, req=None):
    """Can be used with xpartial to register views that simply redirect.

    >>> view = xpartial(pyramid.httpexceptions.HTTPFound, lambda req: req.route_url('.'))
    """
    # workaround for https://github.com/Pylons/pyramid/issues/1428
    req = req or ctx
    if callable(location):
        location = location(req)
    raise cls(location=location)


def gone(ctx, req):
    """view callable."""
    raise pyramid.httpexceptions.HTTPGone()


def view(interface, ctx, req, getadapters=False):
    """Render a resource as pyramid response.

    Using the most appropriate adapter for the accept header sent.

    :param getadapters: If True, the adapter used to render the response and the list of\
    all available adapters for the context are returned as well.
    """
    #
    # if req.matched_route.name.endswith('_alt') -> add rel="canonical" header
    # else: add rel="alternate" header
    #
    adapter, adapters = get_adapter(
        interface, ctx, req, ext=req.matchdict and req.matchdict.get('ext'), getall=True)
    if not adapter:
        raise pyramid.httpexceptions.HTTPNotAcceptable()
    res = adapter.render_to_response(ctx, req)
    if getadapters:
        return res, adapter, adapters
    return res  # pragma: no cover


def _add_link_header(response, url, adapter=None, rel="canonical", mimetype="text/html"):
    if adapter:
        rel = adapter.rel
        mimetype = adapter.send_mimetype or adapter.mimetype
    if mimetype and rel:
        response.headerlist.append(
            ('Link', '<%s>; rel="%s"; type="%s"' % tuple(map(str, [url, rel, mimetype]))))


def index_view(ctx, req):
    if req.is_xhr and 'sEcho' in req.params:
        return datatable_xhr_view(ctx, req)
    res, current, adapters = view(IIndex, ctx, req, getadapters=True)
    if req.matched_route:
        if req.matched_route.name.endswith('_alt'):
            # add the canonical link:
            _add_link_header(res, req.route_url(req.matched_route.name[:-4]))
        else:
            for a in [_a for _a in adapters if _a.rel and _a.extension]:
                _add_link_header(
                    res,
                    req.route_url(req.matched_route.name + '_alt', ext=a.extension),
                    adapter=a)
    return res


def resource_view(ctx, req):
    res, _, adapters = view(IRepresentation, ctx, req, getadapters=True)
    if req.matched_route:
        if req.matched_route.name.endswith('_alt'):
            _add_link_header(res, req.resource_url(ctx))
        else:
            for a in adapters:
                if a.rel and a.extension:
                    _add_link_header(
                        res, req.resource_url(ctx, ext=a.extension), adapter=a)
    return res


def datatable_xhr_view(ctx, req):
    # call get_query, thereby - as side effect - making sure, the counts are set.
    items = ctx.get_query()
    if hasattr(ctx, 'row_class'):
        data = []
        for item in items:
            _d = {'DT_RowId': 'row_%s' % item.pk}
            row_class = ctx.row_class(item)
            if row_class:
                _d['DT_RowClass'] = row_class
            for i, col in enumerate(ctx.cols):
                _d[str(i)] = col.format(item)
            data.append(_d)
    else:
        data = [[col.format(item) for col in ctx.cols] for item in items]

    # sEcho parameter.
    # Note that it strongly recommended for security reasons that you 'cast' this
    # parameter to an integer in order to prevent Cross Site Scripting (XSS) attacks.
    try:
        echo = int(req.params['sEcho'])
    except ValueError:
        echo = 1

    res = {
        "aaData": data,
        "sEcho": str(echo),
        "iTotalRecords": ctx.count_all,
        "iTotalDisplayRecords": ctx.count_filtered,
    }
    return render_to_response('json', res, request=req)


def js(req):
    param_pattern = re.compile(r'\{(?P<name>[a-z]+)(\:[^\}]+)?\}')

    res = [
        "CLLD.base_url = %s;" % json.dumps(req.application_url),
        "CLLD.query_params = %s;" % json.dumps(req.query_params),
    ]
    for route in req.registry.getUtility(IRoutesMapper).get_routes():
        pattern = param_pattern.sub(lambda m: '{%s}' % m.group('name'), route.pattern)
        res.append('CLLD.routes[%s] = %s;' % tuple(map(json.dumps, [route.name, pattern])))
    return Response('\n'.join(res), content_type="text/javascript")


def select_combination(ctx, req):
    if 'parameters' in req.params:
        ids = req.params.getall('parameters')
        if len(ids) > 1:
            id_ = Combination.delimiter.join(ids)
        else:
            id_ = Combination.delimiter.join(req.params['parameters'].split(','))
        raise pyramid.httpexceptions.HTTPFound(req.route_url('combination', id=id_))
    raise pyramid.httpexceptions.HTTPNotFound


def _raise(req):
    """view callable to test error reporting in running apps."""
    raise ValueError('test')


def _ping(req):
    """view to test aliveness of apps."""
    return {'status': 'ok'}


def unapi(req):
    """View callable implementing the server side of the unAPI spec."""
    id_ = req.params.get('id')
    if id_:
        obj = req.ctx_for_url(id_)
        if not obj:
            return pyramid.httpexceptions.HTTPNotFound()

    format_ = req.params.get('format')
    if not format_:
        kw = {'content_type': 'application/xml'}
        if id_:
            formats = [a for n, a in get_adapters(IMetadata, obj, req)]
            kw['status'] = '300 Multiple Choices'
        else:
            formats = []
        body = render(
            'unapiformats.mako',
            {'formats': formats, 'identifier': id_},
            request=req)
        return Response(body, **kw)

    if not id_:
        return pyramid.httpexceptions.HTTPNotFound()

    adapter = None
    for _n, _a in get_adapters(IMetadata, obj, req):
        if _a.unapi_name == format_:
            adapter = _a
            break
    if not adapter:
        return pyramid.httpexceptions.HTTPNotAcceptable()
    return pyramid.httpexceptions.HTTPFound(
        location=req.resource_url(obj, ext=adapter.extension))


#
# The following code implements a view to show a map with parameters from distinct
# datasets. It may be used by CrossGram at some point.
#
class ParameterMultiSelect(MultiSelect):  # pragma: no cover

    """Experimental."""

    def __init__(self, req, name, eid, collection=None, url=None, selected=None):
        MultiSelect.__init__(self, req, name, eid, url='x')
        self.data = []
        for app, rm in req.registry.settings.get('clld.parameters', {}).items():
            for param in rm['resources']:
                self.data.append({
                    'id': '%s-%s' % (app, param['id']),
                    'text': '%s %s: %s' % (app, param['id'], param['name'])})
        self._datadict = dict((d['id'], d) for d in self.data)

    def format_result(self, obj):
        return obj

    def get_urls(self):
        for i, param in enumerate(self.req.params.get('parameters', '').split(',')):
            if param in self._datadict:
                if self.selected is None:
                    self.selected = []
                self.selected.append(self._datadict[param])
                app, pid = param.split('-', 1)
                rm = self.req.registry.settings['clld.parameters'][app]
                yield (app, pid, rm['properties']['uri_template'].format(id=pid))

    def get_default_options(self):
        return {
            'placeholder': "Select parameter",
            'width': 'off',
            'class': 'span6',
            'multiple': True,
            'data': self.data}



from sqlalchemy import Column, Integer, Boolean, Date, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin,
    HasSourceNotNullMixin, Contributor)

__all__ = ('Contribution', 'ContributionReference', 'ContributionContributor')


class Contribution_data(Base, DataMixin):
    pass


class Contribution_files(Base, FilesMixin):
    pass


@implementer(interfaces.IContribution)
class Contribution(Base,
                   PolymorphicBaseMixin,
                   IdNameDescriptionMixin,
                   HasDataMixin,
                   HasFilesMixin):

    """A set of data contributed within the same context by the same contributors."""

    __table_args__ = (UniqueConstraint('name'),)

    date = Column(Date)

    @property
    def primary_contributors(self):
        return [assoc.contributor for assoc in
                sorted(self.contributor_assocs,
                       key=lambda a: (a.ord, a.contributor.id)) if assoc.primary]

    @property
    def secondary_contributors(self):
        return [assoc.contributor for assoc in
                sorted(self.contributor_assocs,
                       key=lambda a: (a.ord, a.contributor.id)) if not assoc.primary]

    def formatted_contributors(self):
        contribs = [' and '.join(c.name for c in self.primary_contributors)]
        if self.secondary_contributors:
            contribs.append(' and '.join(c.name for c in self.secondary_contributors))
        return ' with '.join(contribs)


class ContributionReference(Base, HasSourceNotNullMixin):

    """Association table."""

    __table_args__ = (
        UniqueConstraint('contribution_pk', 'source_pk', 'description'),
    )

    contribution_pk = Column(Integer, ForeignKey('contribution.pk'), nullable=False)
    contribution = relationship(Contribution, innerjoin=True, backref="references")


class ContributionContributor(Base, PolymorphicBaseMixin):

    """Many-to-many association between contributors and contributions."""

    __table_args__ = (UniqueConstraint('contribution_pk', 'contributor_pk'),)

    contribution_pk = Column(Integer, ForeignKey('contribution.pk'), nullable=False)
    contributor_pk = Column(Integer, ForeignKey('contributor.pk'), nullable=False)

    # contributors are ordered.
    ord = Column(Integer, default=1)

    # we distinguish between primary and secondary (a.k.a. 'with ...') contributors.
    primary = Column(Boolean, default=True)

    contribution = relationship(
        Contribution, innerjoin=True, backref='contributor_assocs')
    contributor = relationship(
        Contributor, innerjoin=True, lazy=False, backref='contribution_assocs')



from sqlalchemy import func, Index


def tsvector(obj):  # pragma: no cover
    return func.to_tsvector('english', '{0}'.format(obj))


def index(name, col, bind):  # pragma: no cover
    Index(name, col, postgresql_using='gin').create(bind)


def search(col, qs):  # pragma: no cover
    # https://bitbucket.org/zzzeek/sqlalchemy/issues/3160/postgresql-to_tsquery-docs-and
    query = func.plainto_tsquery('english', qs)
    return col.op('@@')(query)



from clld.db.models import common
from clld.web.datatables.unitparameter import Unitparameters

from testutils import handle_dt


def test_table(env):
    handle_dt(env['request'], Unitparameters, common.UnitParameter)



from sqlalchemy.orm import joinedload
from clld.web import datatables
from clld.web.datatables.base import LinkCol, Col, LinkToMapCol

from {{cookiecutter.directory_name}} import models


def includeme(config):
    """register custom datatables"""



"""Adapter registry to be included by pyramid configurator."""
import os

from pyramid.path import AssetResolver

from clld import interfaces
from clld.web.adapters.base import Index, Representation, Json
from clld.web.adapters.geojson import (
    GeoJson, GeoJsonLanguages, GeoJsonParameter, GeoJsonParameterFlatProperties,
)
from clld.web.adapters.md import BibTex, TxtCitation
from clld.web.adapters.rdf import Rdf, RdfIndex
from clld.web.adapters import biblio
from clld.lib.rdf import FORMATS as RDF_NOTATIONS


def template_exists(config, relpath):
    asset_resolver = AssetResolver()
    for md in config.registry.settings['mako.directories']:
        asset_descriptor = asset_resolver.resolve('/'.join([md, relpath]))
        if os.path.exists(asset_descriptor.abspath()):
            return True
    return False


def register_resource_adapters(config, rsc):
    name, interface = rsc.name, rsc.interface

    cls = type('Json%s' % rsc.model.__name__, (Json,), {})
    config.register_adapter(
        cls, interface, to_=interfaces.IRepresentation, name=Json.mimetype)

    specs = []

    if rsc.with_index:
        specs.append((Index, 'application/atom+xml', 'atom', 'index_atom.mako', {}))
        if template_exists(config, name + '/index_html.mako'):
            # ... as html index
            specs.append((Index, 'text/html', 'html', name + '/index_html.mako', {}))

    # ... as RDF in various notations
    rdf_resource_template = name + '/rdf.mako'
    if not template_exists(config, rdf_resource_template):
        rdf_resource_template = 'resource_rdf.mako'

    for notation in RDF_NOTATIONS.values():
        specs.append((
            Rdf,
            notation.mimetype,
            notation.extension,
            rdf_resource_template,
            {'name': 'RDF serialized as %s' % notation.name, 'rdflibname': notation.name},
        ))

    # ... as RDF collection index
    rdf_xml = RDF_NOTATIONS['xml']
    specs.append((
        RdfIndex,
        rdf_xml.mimetype,
        rdf_xml.extension,
        'index_rdf.mako',
        {'rdflibname': rdf_xml.name}))

    if template_exists(config, name + '/detail_html.mako'):
        # ... as html details page
        specs.append(
            (Representation, 'text/html', 'html', name + '/detail_html.mako', {}))
    if template_exists(config, name + '/snippet_html.mako'):
        # ... as html snippet (if the template exists)
        specs.append((
            Representation,
            'application/vnd.clld.snippet+xml',
            'snippet.html',
            name + '/snippet_html.mako',
            {'rel': None}))

    config.register_adapters([[interface] + list(spec) for spec in specs])


def includeme(config):
    """register adapters."""
    specs = []

    # citeable resources are available as html page listing available metadata formats:
    for _if in [interfaces.IContribution, interfaces.IDataset]:
        specs.append((
            _if,
            Representation,
            'application/vnd.clld.md+xml',
            'md.html',
            'md_html.mako',
            {'rel': 'describedby'}))

    specs.append((
        interfaces.ILanguage, Index,
        'application/vnd.google-earth.kml+xml',
        'kml',
        'clld:web/templates/language/kml.mako',
        {'send_mimetype': 'application/xml'}))

    config.register_adapters(specs)

    for cls in [BibTex, TxtCitation]:
        for if_ in [interfaces.IRepresentation, interfaces.IMetadata]:
            for adapts in [interfaces.IContribution, interfaces.IDataset]:
                config.register_adapter(cls, adapts, if_, name=cls.mimetype)

    config.register_adapter(
        GeoJsonLanguages,
        interfaces.ILanguage,
        interfaces.IIndex,
        name=GeoJson.mimetype)
    config.register_adapter(
        GeoJsonParameter,
        interfaces.IParameter,
        name=GeoJson.mimetype)
    config.register_adapter(
        GeoJsonParameterFlatProperties,
        interfaces.IParameter,
        name=GeoJsonParameterFlatProperties.mimetype)
    config.include(biblio)


def get_adapters(interface, ctx, req):
    # ctx can be a DataTable instance. In this case we create a resource by instantiating
    # the model class associated with the DataTable
    resource = ctx.model() if hasattr(ctx, 'model') else ctx
    return req.registry.getAdapters([resource], interface)


def get_adapter(interface, ctx, req, ext=None, name=None, getall=False):
    """Retrieve matching adapter.

    :param interface: Interface class to lookup adapter for.
    """
    adapters = dict(get_adapters(interface, ctx, req))

    if not ext and not name and (
        not req.accept or ('*/*' in str(req.accept) and 'q=' not in str(req.accept))
    ):
        # force text/html in case there are no specific criteria to decide
        # or we suspect some weird IE accept header.
        # see also http://www.gethifi.com/blog/browser-rest-http-accept-headers
        ext = 'html'

    if ext:
        # find adapter by requested file extension
        adapter = [r for r in adapters.values() if r.extension == ext]
    elif name:
        # or by mime type
        adapter = adapters.get(name)
    else:
        # or by content negotiation
        if not hasattr(req.accept, 'acceptable_offers'):  # pragma: no cover
            # pre WebOb 1.8 (should be dropped once pyramid requires WebOb>=1.8):
            adapter = adapters.get(req.accept.best_match(adapters.keys()))
        else:
            try:
                adapter = adapters.get(req.accept.acceptable_offers(list(adapters.keys()))[0][0])
            except IndexError:
                adapter = None

    res = None
    if isinstance(adapter, list):
        if adapter:
            res = adapter[0]
    else:
        res = adapter
    if getall:
        res = (res, adapters.values())
    return res



# flake8: noqa
from clld.db.meta import (Base, PolymorphicBaseMixin, CustomModelMixin)

from ._mixins import (IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin)

from .config import Config

from .language import (Language_data, Language_files, Language,
    LanguageSource,
    IdentifierType, Identifier, LanguageIdentifier)

from .gloss import GlossAbbreviation

from .source import (Source_data, Source_files, Source, HasSourceMixin,
    HasSourceNotNullMixin)

from .contributor import Contributor_data, Contributor_files, Contributor

from .contribution import (Contribution_data, Contribution_files, Contribution,
    ContributionReference, ContributionContributor)

from .dataset import Dataset_data, Dataset_files, Dataset

from .sentence import Sentence_data, Sentence_files, Sentence, SentenceReference

from .valueset import ValueSet_data, ValueSet_files, ValueSet, ValueSetReference

from .value import Value_data, Value_files, Value, ValueSentence

from .parameter import (DomainElement_data, DomainElement_files, DomainElement,
    Parameter_data, Parameter_files, Parameter,
    CombinationDomainElement, Combination)

from .unit import Unit_data, Unit_files, Unit

from .unitvalue import UnitValue_data, UnitValue_files, UnitValue

from .unitparameter import (UnitDomainElement_data, UnitDomainElement_files,
    UnitDomainElement,
    UnitParameter_data, UnitParameter_files, UnitParameter)



"""
Functionality to configure select2 (multi)select widgets.

.. seealso:: http://ivaynberg.github.io/select2/
"""
from clld.db.meta import DBSession
from clld.db.models.common import Parameter
from clld.web.util.helpers import JS
from clld.web.util.component import Component

__all__ = ['MultiSelect', 'CombinationMultiSelect']


class MultiSelect(Component):

    """A select component based on select2."""

    __template__ = 'clld:web/templates/multiselect.mako'

    def __init__(self,
                 req,
                 name,
                 eid,
                 collection=None,
                 url=None,
                 selected=None,
                 multiple=True):
        """Initialize.

        for selections using remote data pass url, otherwise pass collection to force
        creation of a select element (with no order preserved for the selection) or
        specify a 'data' member in options.

        :param selected: list of items to be marked as selected or None.
        """
        self.req = req
        self.collection = collection
        self.url = url
        self.eid = eid
        self.name = name
        self.selected = selected
        self.multiple = multiple

    def get_default_options(self):
        res = {
            'placeholder': "Search %s" % self.name,
            'width': 'element',
        }
        if self.url:
            res.update(
                minimumInputLength=2,
                multiple=self.multiple,
                ajax={
                    'url': self.url,
                    'dataType': 'json',
                    'data': JS('CLLD.MultiSelect.data'),
                    'results': JS('CLLD.MultiSelect.results'),
                }
            )
        return res

    def format_result(self, obj):
        """called for each matching result.

        :return: dict which can be serialized as JSON for use by the select2 component.
        """
        return {
            'id': getattr(obj, 'id', obj.pk),
            'text': '%s' % getattr(obj, 'label', obj)}

    def render(self, selected=None):
        """allow the list of selected items to be specified upon rendering, too."""
        if selected:
            self.selected = selected
        return Component.render(self)


class CombinationMultiSelect(MultiSelect):

    """Multiple selection of parameters for combination.

    >>> ms = CombinationMultiSelect(None)
    """

    def __init__(self, req, name='parameters', eid='ms-parameters', combination=None,
                 **kw):
        if combination:
            kw['selected'] = combination.parameters
        MultiSelect.__init__(self, req, name, eid, **kw)

    @classmethod
    def query(cls):
        return DBSession.query(Parameter)

    def format_result(self, obj):
        return {'id': obj.id, 'text': '%s: %s' % (obj.id, obj.name)}

    def get_options(self):
        return {
            'data': [self.format_result(p) for p in self.query()],
            'multiple': True,
            'maximumSelectionSize': 4}



from pathlib import Path

from clld.web.assets import environment

import {{cookiecutter.directory_name}}


environment.append_path(
    Path({{cookiecutter.directory_name}}.__file__).parent.joinpath('static').as_posix(),
    url='/{{cookiecutter.directory_name}}:static/')
environment.load_path = list(reversed(environment.load_path))



import itertools

import pytest

from clld.lib.rdf import *


@pytest.mark.parametrize(
    'name,url',
    [
        ('dcterms:title', 'http://purl.org/dc/terms/title'),
        ('noprefix:lname', 'noprefix:lname'),
        ('rdf:nolname', 'rdf:nolname'),
        ('nocolon', 'nocolon'),
    ]
)
def test_expand_prefix(name, url):
    assert str(expand_prefix(name)) == url


def test_url_for_qname():
    assert url_for_qname('dcterms:title') == 'http://purl.org/dc/terms/title'


@pytest.mark.parametrize(
    'subject,props,substring',
    [
        ('subject', [('foaf:name', 'a name'), ('foaf:homepage', 'http://example.org')], None),
        ('subject', [('rdf:about', 'x'), ('foaf:homepage', 'http://example.org')], None),
        ('subject', [('about', 'x'), ('foaf:homepage', 'http://example.org')], None),
        ('http://example.org', [('dcterms:title', 'ttt')], 'ttt'),
    ]
)
def test_properties_as_xml_snippet(subject,props,substring):
    p = properties_as_xml_snippet(subject, props)
    assert not substring or (substring in p)


@pytest.mark.parametrize(
    'from_,to_',
    [
        (f1, f2) for f1, f2 in
        itertools.product(list(FORMATS.keys()) + [None], repeat=2)
        if f1 not in {None, 'nt'} and f2 != 'nt'
    ]
)
def test_convert(from_, to_):
    convert(ClldGraph().serialize(format=from_), from_, to_)



from sqlalchemy import Column, Integer, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces

from . import IdNameDescriptionMixin, DataMixin, HasDataMixin, FilesMixin, HasFilesMixin

__all__ = ('UnitDomainElement', 'UnitParameter')


class UnitDomainElement_data(Base, DataMixin):
    pass


class UnitDomainElement_files(Base, FilesMixin):
    pass


@implementer(interfaces.IUnitDomainElement)
class UnitDomainElement(Base,
                        PolymorphicBaseMixin,
                        IdNameDescriptionMixin,
                        HasDataMixin,
                        HasFilesMixin):

    """Domain element for the domain of a UnitParameter."""

    __table_args__ = (
        UniqueConstraint('unitparameter_pk', 'name'),
        UniqueConstraint('unitparameter_pk', 'ord'),
    )

    unitparameter_pk = Column(Integer, ForeignKey('unitparameter.pk'), nullable=False)
    ord = Column(Integer)

    def url(self, request):
        return request.resource_url(self.parameter, _anchor='DE-%s' % self.id)


class UnitParameter_data(Base, DataMixin):
    pass


class UnitParameter_files(Base, FilesMixin):
    pass


@implementer(interfaces.IUnitParameter)
class UnitParameter(Base,
                    PolymorphicBaseMixin,
                    IdNameDescriptionMixin,
                    HasDataMixin,
                    HasFilesMixin):

    """A measurable attribute of a unit."""

    domain = relationship(
        'UnitDomainElement', backref='parameter', order_by=UnitDomainElement.id)



"""Default DataTable for ValueSet objects."""
import functools

from sqlalchemy.orm import joinedload

from clld.db.models.common import (
    ValueSet, Parameter, Language, Contribution, ValueSetReference,
)
from clld.web.datatables.base import (
    DataTable, LinkCol, DetailsRowLinkCol, LinkToMapCol, RefsCol,
)


class Valuesets(DataTable):

    """Default DataTable for ValueSet objects."""

    __constraints__ = [Parameter, Contribution, Language]

    def base_query(self, query):
        query = query.join(Language)\
            .options(
                joinedload(
                    ValueSet.language
                ),
                joinedload(
                    ValueSet.references
                ).joinedload(
                    ValueSetReference.source
                ))

        if self.language:
            query = query.join(Parameter).options(joinedload(ValueSet.parameter))
            return query.filter(ValueSet.language_pk == self.language.pk)

        if self.parameter:
            return query.filter(ValueSet.parameter_pk == self.parameter.pk)

        if self.contribution:
            query = query.join(Parameter)
            return query.filter(ValueSet.contribution_pk == self.contribution.pk)

        return query

    def col_defs(self):
        refs_col = RefsCol(self, 'references')
        res = [DetailsRowLinkCol(self, 'd')]

        def get(what, i):
            return getattr(i, {'p': 'parameter', 'l': 'language'}[what])

        if self.parameter:
            return res + [
                LinkCol(self, 'language',
                        model_col=Language.name, get_obj=functools.partial(get, 'l')),
                refs_col,
                LinkToMapCol(self, 'm', get_obj=functools.partial(get, 'l')),
            ]

        if self.language:
            return res + [
                LinkCol(self, 'parameter',
                        model_col=Parameter.name, get_obj=functools.partial(get, 'p')),
                refs_col,
            ]

        return res + [
            LinkCol(self, 'language', model_col=Language.name, get_obj=functools.partial(get, 'l')),
            LinkCol(
                self, 'parameter', model_col=Parameter.name, get_obj=functools.partial(get, 'p')),
            refs_col,
        ]

    def toolbar(self):
        return ''



from markupsafe import Markup

from clld.interfaces import IDataTable
from clld.web.util.component import Component
from clld.web.util.htmllib import HTML, literal


class DownloadWidget(Component):

    """Button group, grouping an info popover and a download selection."""

    def __init__(self, req, ctx, obj, dl_url_tmpl, interface, **kw):
        self.req = req
        self.ctx = ctx
        self.obj = obj
        self.dl_url_tmpl = dl_url_tmpl
        self.interface = interface
        kw.setdefault('doc_position', 'right')
        kw.setdefault('exclude', ['html', 'snippet.html'])
        self._kw = kw
        self._opener_class = 'dl-widget-%s' % self.interface.__name__
        self._id_prefix = 'dt-dl-' if IDataTable.providedBy(ctx) else 'rsc-dl-'

    def get_default_options(self):
        return self._kw

    def doc(self):
        """Override to supply additional information in the information popup.

        :return: HTML string.
        """
        return ''

    def js(self):
        return Markup(HTML.script(literal("""\
    $(document).ready(function() {
        $('.%s').clickover({
            html: true,
            title: 'Download information',
            placement: '%s',
            trigger: 'click'
        });
    });""" % (self._opener_class, self.options['doc_position']))))

    def dl_link(self, adapter):
        return HTML.a(
            adapter.name or adapter.extension,
            href="#",
            id=self._id_prefix + adapter.extension,
            onclick="document.location.href = %s; return false;"
                    % (self.dl_url_tmpl % adapter.extension))

    def render(self, no_js=False):
        adapters = [a for n, a in
                    self.req.registry.getAdapters([self.obj], self.interface)
                    if a.extension not in set(self.options['exclude'])]
        adapters = sorted(adapters, key=lambda x: x.extension)
        adoc = []
        for adapter in adapters:
            if adapter.__doc__:
                adoc.append(HTML.dt(adapter.name or adapter.extension))
                adoc.append(HTML.dd(adapter.__doc__))
        doc = HTML.div(
            self.doc(),
            HTML.p(
                """You may download alternative representations of the data on
"%s" by clicking the button """ % self.ctx,
                HTML.i(class_='icon-download-alt')),
            HTML.dl(*adoc))
        res = HTML.div(
            HTML.button(
                HTML.i(class_='icon-info-sign icon-white'),
                class_='btn btn-info %s' % self._opener_class,
                **{'data-content': str(doc), 'type': 'button'}),
            HTML.a(
                HTML.i(class_='icon-download-alt'),
                HTML.span(class_="caret"),
                **{
                    'class_': "btn dropdown-toggle",
                    'data-toggle': "dropdown",
                    'href': "#",
                    'id': self._id_prefix + "opener"}),
            HTML.ul(
                *[HTML.li(self.dl_link(adapter)) for adapter in adapters],
                **dict(class_="dropdown-menu")),
            class_='btn-group right')
        if no_js:  # pragma: no cover
            return res
        return HTML.div(res, self.js())



"""
Create the skeleton for a clld app project.

Variables:
- directory_name: The name of the project directory. This will also be used as name
  of the python package.
- cldf_module: If the app data is initialized from a CLDF dataset, specify the CLDF
  module this dataset conforms to (Wordlist|StructureDataset|Dictionary|Generic).
  Leave empty otherwise.
  Note that this requires passing an `--cldf` option to `clld initdb`.
- mpg: Specify "y" if the app is served from MPG servers, and thus needs to fulfill
  certain legal obligations (n|y).
"""
import pathlib

from clldutils.clilib import PathType
import clld

try:
    from cookiecutter.main import cookiecutter
    from cookiecutter.exceptions import OutputDirExistsException
except ImportError:  # pragma: no cover
    cookiecutter = None


def kv(string):
    return string.split('=', maxsplit=1)


def register(parser):
    parser.add_argument(
        'outdir',
        type=PathType(must_exist=False, type='dir'),
        help="Output directory. The last path segment will be used as default value for the"
             "'directory_name' variable."
    )
    parser.add_argument(
        'variables',
        nargs='*',
        type=kv,
        help="If run non-interactively, defaults for the template variables can be passed in"
             "as 'key=value'-formatted arguments",
    )
    parser.add_argument(
        '-f', '--force',
        default=False,
        help="Overwrite an existing project directory",
        action='store_true',
    )
    parser.add_argument(
        '--quiet',
        default=False,
        help="Run non-interactively, i.e. do not prompt for template variable input.",
        action='store_true',
    )


def run(args):
    if not cookiecutter:  # pragma: no cover
        print('cookiecutter must be installed to create project templates')
        return 1

    extra = {'directory_name': str(args.outdir.name)}
    extra.update(dict(args.variables))
    try:
        cookiecutter(
            str(pathlib.Path(clld.__file__).parent / 'project_template'),
            output_dir=str(args.outdir.parent),
            extra_context=extra,
            overwrite_if_exists=args.force,
            no_input=args.quiet,
        )
    except OutputDirExistsException as e:
        print(e)
        print('Run with --force option to overwrite!')
        raise ValueError()



import pytest
from clld.db.models import common
from clld.web.datatables.unit import Units

from testutils import handle_dt


@pytest.mark.parametrize(
    "params",
    [
        {},
        {'language': 'language'},
    ])
def test_Units(request_factory, params):
    with request_factory(params=params) as req:
        handle_dt(req, Units, common.Unit)



from sqlalchemy import Column, Integer, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces

from . import (
    IdNameDescriptionMixin, DataMixin, HasDataMixin, FilesMixin, HasFilesMixin, Language)

__all__ = ('Unit',)


class Unit_data(Base, DataMixin):
    pass


class Unit_files(Base, FilesMixin):
    pass


@implementer(interfaces.IUnit)
class Unit(Base,
           PolymorphicBaseMixin,
           IdNameDescriptionMixin,
           HasDataMixin,
           HasFilesMixin):
    """A linguistic unit of a language."""

    __table_args__ = (UniqueConstraint('language_pk', 'id'),)

    language_pk = Column(Integer, ForeignKey('language.pk'), nullable=False)
    language = relationship(Language, innerjoin=True)



"""This module provides functionality for handling our data as rdf."""
import io
import collections

from clldutils.misc import encoded
from rdflib import Graph, URIRef, Literal
from rdflib.namespace import (
    Namespace, DC, DCTERMS, DOAP, FOAF, OWL, RDF, RDFS, SKOS, VOID, XMLNS, XSD,
)
# make flake8 happy, but still have the following importable from here:
assert DOAP
assert XMLNS

__all__ = ['FORMATS', 'NAMESPACES', 'expand_prefix', 'url_for_qname',
           'ClldGraph', 'properties_as_xml_snippet', 'convert']

Notation = collections.namedtuple('Notation', 'name extension mimetype uri')

FORMATS = dict((n.name, n) for n in [
    Notation('xml', 'rdf', 'application/rdf+xml', 'http://www.w3.org/ns/formats/RDF_XML'),
    Notation('n3', 'n3', 'text/n3', 'http://www.w3.org/ns/formats/N3'),
    Notation('nt', 'nt', 'text/nt', 'http://www.w3.org/ns/formats/N-Triples'),
    Notation('turtle', 'ttl', 'text/turtle', 'http://www.w3.org/ns/formats/Turtle')])

NAMESPACES = {
    "rdf": RDF,
    "void": VOID,
    "foaf": FOAF,
    "frbr": Namespace("http://purl.org/vocab/frbr/core#"),
    "dcterms": DCTERMS,
    "dctype": Namespace("http://purl.org/dc/dcmitype/"),
    "rdfs": RDFS,
    "geo": Namespace("http://www.w3.org/2003/01/geo/wgs84_pos#"),
    "isbd": Namespace("http://iflastandards.info/ns/isbd/elements/"),
    "skos": SKOS,
    "dc": DC,
    "gold": Namespace("http://purl.org/linguistics/gold/"),
    "lexvo": Namespace("http://lexvo.org/ontology#"),
    "vcard": Namespace("http://www.w3.org/2001/vcard-rdf/3.0#"),
    "bibo": Namespace("http://purl.org/ontology/bibo/"),
    "owl": OWL,
    "xsd": XSD,
}


def expand_prefix(p):
    """Expand default prefixes if possible.

    :param p: a qualified name in prefix:localname notation or a URL.
    :return: a string URL or a URIRef
    """
    if isinstance(p, str) and ':' in p:
        prefix, name = p.split(':', 1)
        if prefix in NAMESPACES:
            return getattr(NAMESPACES[prefix], name, p)
    return p


def url_for_qname(qname):
    """Expand qname to full URL respecting our default prefixes."""
    return str(expand_prefix(qname))


class ClldGraph(Graph):

    """Augmented rdflib.Graph.

    Augment the standard rdflib.Graph by making sure our standard ns prefixes are
    always bound.
    """

    def __init__(self, *args, **kw):
        super(ClldGraph, self).__init__(*args, **kw)
        for prefix, ns in NAMESPACES.items():
            self.bind(prefix, ns)


def properties_as_xml_snippet(subject, props):
    """Serialize props of subject as RDF-XML snippet."""
    if isinstance(subject, str):
        subject = URIRef(subject)
    g = ClldGraph()
    if props:
        for p, o in props:
            p = expand_prefix(p)
            if p:
                if isinstance(o, str):
                    if o.startswith('http://') or o.startswith('https://'):
                        o = URIRef(o)
                    else:
                        o = Literal(o)
                try:
                    g.add((subject, p, o))
                except AssertionError:
                    pass
    res = []
    in_desc = False
    xml = g.serialize(format='xml')
    if not isinstance(xml, str):
        xml = xml.decode('utf8')  # pragma: no cover
    for line in xml.split('\n'):
        if line.strip().startswith('</rdf:Description'):
            break
        if in_desc:
            res.append(line)
        if line.strip().startswith('<rdf:Description'):
            in_desc = True
    return '\n'.join(res)


def convert(string, from_, to_):
    if from_ == to_:
        return encoded(string)
    assert from_ in FORMATS and (to_ is None or to_ in FORMATS)
    g = Graph()
    g.parse(io.BytesIO(encoded(string)), format=from_)
    if to_ is None:
        return g
    out = io.BytesIO()
    g.serialize(out, format=to_)
    out.seek(0)
    return out.read()



import pytest
from clld.db.models import common
from clld.web.maps import (
    Map, ParameterMap, LanguageMap, SelectedLanguagesMap, CombinationMap, FilterLegend,
    Layer,
)


@pytest.mark.parametrize(
    "route,params,map_cls,ctx_cls,ctx_id",
    [
        ('language', dict(z=3, lat=1, lng=1), Map, common.Parameter, 'parameter'),
        ('language', {'z': 'abc'}, Map, common.Parameter, 'parameter'),
        ('parameter', {}, ParameterMap, common.Parameter, 'parameter'),
        ('parameter', {}, ParameterMap, common.Parameter, 'no-domain'),
        ('language', {}, LanguageMap, common.Language, 'language'),
    ])
def test_Map(request_factory, route, params, map_cls, ctx_cls, ctx_id):
    with request_factory(matched_route=route, params=params) as req:
        m = map_cls(ctx_cls.get(ctx_id), req)
        m.render()


def test_SelectedLanguagesMap(env):
    m = SelectedLanguagesMap(None, env['request'], [common.Language.first()])
    m.render()


def test_CombinationMap(env):
    ctx = common.Combination(common.Parameter.first())
    assert ctx.domain
    ctx.multiple = [common.Language.first()]
    dt = CombinationMap(ctx, env['request'])
    dt.render()


def test_layers(env):
    class TestMap(Map):
        def get_layers(self):
            yield Layer('l1', 'ln1', [], representation=555)
            yield Layer('l2', 'ln2', [], representation=333)

    m = TestMap(None, env['request'])
    assert '888' in m.render()


def test_FilterLegend(request_factory):
    from clld.web.datatables import Languages

    class FLanguages(Languages):
        def col_defs(self):
            cols = Languages.col_defs(self)
            cols[1].choices = ['name']
            return cols

    class FMap(Map):
        def get_legends(self):
            yield FilterLegend(
                self,
                '',
                col='name',
                dt=FLanguages(self.req, common.Language))

    with request_factory(matched_route='language') as req:
        map_ = FMap(common.Language.first(), req)
        map_.render()



"""

"""
import pathlib
import contextlib

import transaction
from clldutils import db
from clldutils.clilib import PathType
from clld.cliutil import SessionContext, BootstrappedAppConfig
try:
    from pycldf import Dataset
except ImportError:  # pragma: no cover
    Dataset = None

try:
    from cldfcatalog import Catalog
except ImportError:  # pragma: no cover
    Catalog = None


def register(parser):
    parser.add_argument(
        "config-uri",
        action=BootstrappedAppConfig,
        help="ini file providing app config",
    )
    parser.add_argument(
        '--prime-cache-only',
        action='store_true',
        default=False,
    )
    parser.add_argument(
        '--cldf',
        type=PathType(type='file'),
        default=None,
    )
    for name in ['concepticon', 'glottolog']:
        parser.add_argument(
            '--' + name,
            metavar=name.upper(),
            help='Path to repository clone of {0} data'.format(name.capitalize()),
            default=None)
        parser.add_argument(
            '--{0}-version'.format(name),
            help='Version of {0} data to checkout'.format(name.capitalize()),
            default=None)


def run(args):
    if (args.glottolog or args.concepticon) and Catalog is None:  # pragma: no cover
        print('To use reference catalogs you must install the cldfcatalog package!')
        return 10

    if args.cldf:  # pragma: no cover
        if not Dataset:
            print('To import CLDF data you must install the pycldf package!')
            return 10
        args.cldf = Dataset.from_metadata(args.cldf)

    with contextlib.ExitStack() as stack:
        if not args.prime_cache_only:
            stack.enter_context(db.FreshDB.from_settings(args.settings, log=args.log))
        stack.enter_context(SessionContext(args.settings))

        for name in ['concepticon', 'glottolog']:
            if getattr(args, name):  # pragma: no cover
                if getattr(args, name + '_version'):
                    stack.enter_context(
                        Catalog(getattr(args, name), tag=getattr(args, name + '_version')))
                else:
                    setattr(args, name, pathlib.Path(getattr(args, name)))

        if not args.prime_cache_only:
            with transaction.manager:
                if args.initializedb:  # pragma: no cover
                    args.initializedb.main(args)
        if hasattr(args.initializedb, 'prime_cache'):
            with transaction.manager:  # pragma: no cover
                args.initializedb.prime_cache(args)



"""Default DataTable for UnitParameter objects."""
from clld.web.datatables.base import DataTable, LinkCol


class Unitparameters(DataTable):

    """Default DataTable for UnitParameter objects."""

    def col_defs(self):
        return [LinkCol(self, 'name')]



from clld.util import *


def test_safe_overwrite(tmp_path):
    target = tmp_path / 'a' / 'b'
    with safe_overwrite(target) as tmp:
        tmp.write_text('stuff', encoding='utf8')

    assert not tmp.exists()
    assert target.exists()

    with safe_overwrite(target) as tmp:
        tmp.write_text('other', encoding='utf8')

    with target.open(encoding='utf8') as fp:
        assert fp.read() == 'other'


def test_random_string():
    assert len(random_string(5)) == 5


def test_summary():
    # text consisting of unique words
    text = "This is a long text, which we want to summarize."

    assert summary(text[:20]) == text[:20]
    assert summary(text, len(text) - 1).endswith('...')
    assert summary('One verylongword', 10) == 'One ...'
    assert summary('One verylongword', 2) == '...'


def test_DeclEnum():
    class A(DeclEnum):
        val1 = '1', 'value 1'
        val2 = '2', 'value 2'

    db_type = A.db_type()
    assert A.val1 == db_type.process_result_value(
        db_type.process_bind_param(A.val1, None), None)
    assert db_type.process_result_value(
        db_type.process_bind_param(None, None), None) is None
    assert A.val1.__json__() == str(A.val1)



"""Functionality to format HTML pagination controls."""
from clldutils.paginate import Page

from clld.web.util.component import Component


class Pager(Component):

    """A pager component based on paginate.Page.

    >>> pager = Pager(None, range(100), page=4, url_maker=lambda p: 'page %s' % p)
    >>> assert pager.render()
    """

    def __init__(self, req, collection, **kw):
        self.req = req
        self.page = Page(collection, **kw)

    def render(self):
        html = self.page.pager(
            format='<div class="pagination"><ul><li>~3~</li></ul></div>',
            separator='</li><li>',
            curpage_attr={'class': 'active'},
            dotdot_attr={'class': 'disabled'})
        return html\
            .replace('<span ', '<a ')\
            .replace('</span>', '</a>')\
            .replace('<li><a class="active"', '<li class="active"><a')\
            .replace('<li><a class="disabled"', '<li class="disabled"><a')



import logging

import pytest

from clld.__main__ import main


@pytest.mark.filterwarnings("ignore:No module named")
def test_create_downloads(testsdir):
    main(['create_downloads', '-l', str(testsdir / 'test.ini'), 'abc'],
         log=logging.getLogger(__name__))


def test_create(tmp_path):
    outdir = tmp_path / 'app'
    main(['create', '--quiet', str(outdir)])
    assert outdir.joinpath('setup.py').exists()
    with pytest.raises(ValueError):
        main(['create', '--quiet', str(outdir)])
    main(['create', '--quiet', '--force', str(outdir), "domain=example.org"])
    assert "example.org" in outdir.joinpath(
        'app', 'scripts', 'initializedb.py').read_text(encoding='utf8')


@pytest.mark.filterwarnings("ignore:No module named")
def test_initdb(tmp_path):
    tmp_path.joinpath('tests').mkdir()
    cfg = tmp_path / 'tests' / 'test.ini'
    cfg.write_text("""\
[app:main]
use = call:testutils:main
sqlalchemy.url = sqlite:///{}
    """.format(tmp_path / 'db.sqlite'), encoding='utf8')
    main(['initdb', str(cfg)])

    with pytest.raises(ValueError):
        main(['initdb', str(tmp_path / 'xyz.ini')])



"""Default DataTable for Parameter objects."""
from clld.web.datatables.base import DataTable, LinkCol, DetailsRowLinkCol


class Parameters(DataTable):

    """Default DataTable for Parameter objects."""

    def col_defs(self):
        return [
            DetailsRowLinkCol(self, 'd'),
            LinkCol(self, 'name'),
        ]



from sqlalchemy import (
    Column,
    Float,
    Integer,
    String,
    Unicode,
    CheckConstraint,
    UniqueConstraint,
    ForeignKey,
)
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.associationproxy import association_proxy

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces
from clld.util import DeclEnum

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin)

__all__ = (
    'Language', 'LanguageSource',
    'IdentifierType', 'Identifier', 'LanguageIdentifier',
)


class Language_data(Base, DataMixin):

    """Associated data mapper."""


class Language_files(Base, FilesMixin):

    """Associated files mapper."""


@implementer(interfaces.ILanguage)
class Language(Base,
               PolymorphicBaseMixin,
               IdNameDescriptionMixin,
               HasDataMixin,
               HasFilesMixin):

    """Languages are the main objects of discourse.

    We attach a geo-coordinate to them to be able to put them on maps.
    """

    latitude = Column(
        Float(),
        CheckConstraint('-90 <= latitude and latitude <= 90'),
        doc='geographical latitude in WGS84')
    longitude = Column(
        Float(),
        CheckConstraint('-180 <= longitude and longitude <= 180 '),
        doc='geographical longitude in WGS84')
    identifiers = association_proxy('languageidentifier', 'identifier')

    def get_identifier_objs(self, type_):
        return [i for i in self.identifiers if i.type == getattr(type_, 'value', type_)]

    def get_identifier(self, type_):
        objs = self.get_identifier_objs(type_)
        if objs:
            return objs[0].name

    @property
    def iso_code(self):
        return self.get_identifier(IdentifierType.iso)

    @property
    def glottocode(self):
        return self.get_identifier(IdentifierType.glottolog)


class LanguageSource(Base):

    """Association table."""

    __table_args__ = (UniqueConstraint('language_pk', 'source_pk'),)

    language_pk = Column(Integer, ForeignKey('language.pk'), nullable=False)
    source_pk = Column(Integer, ForeignKey('source.pk'), nullable=False)


class IdentifierType(DeclEnum):

    """Known language identifiers."""

    iso = 'iso639-3', 'ISO 639-3', 'https://iso639-3.sil.org/code/{0.name}'
    wals = 'wals', 'WALS Code', 'http://wals.info/languoid/lect/wals_code_{0.name}'
    glottolog = 'glottolog', 'Glottocode', \
                'http://glottolog.org/resource/languoid/id/{0.name}'
    ethnologue = 'ethnologue', 'Ethnologue', 'http://www.ethnologue.com/language/{0.name}'


class Identifier(Base, IdNameDescriptionMixin):

    """A language identifier.

    We want to be able to link languages to languages in other systems. Thus,
    we store identifiers of various types like 'wals', 'iso639-3', 'glottolog'.
    But we might as well just store alternative names for languages.
    """

    __table_args__ = (UniqueConstraint('name', 'type', 'description', 'lang'),)

    id = Column(String)
    type = Column(String)
    lang = Column(String(3), default='en')

    def url(self):
        try:
            return IdentifierType.from_string(self.type).args[0].format(self)
        except ValueError:
            return


class LanguageIdentifier(Base):

    """Association table.

    Languages are linked to identifiers with an optional description of this
    linkage, e.g. 'is dialect of'.
    """

    __table_args__ = (UniqueConstraint('language_pk', 'identifier_pk'),)

    language_pk = Column(Integer, ForeignKey('language.pk'), nullable=False)
    identifier_pk = Column(Integer, ForeignKey('identifier.pk'), nullable=False)
    description = Column(Unicode)

    identifier = relationship(Identifier, innerjoin=True)
    language = relationship(
        Language, innerjoin=True,
        backref=backref("languageidentifier", cascade="all, delete-orphan"))



"""Various helper functions to be used mainly in templates.

.. note:: This module is available within Mako templates as ``h``.
"""
import os
import re
from itertools import groupby  # we just import this to have it available in templates!
import datetime  # we just import this to have it available in templates!
from base64 import b64encode
from urllib.parse import quote, urlencode, urlparse, urlsplit, urlunsplit, SplitResult

from sqlalchemy import or_
from sqlalchemy.orm import joinedload
from markupsafe import Markup
from pyramid.renderers import render as pyramid_render
from pyramid.threadlocal import get_current_request
from pyramid.interfaces import IRoutesMapper
from zope.interface import providedBy
from clldutils.misc import xmlchars
from clldutils.coordinates import degminsec

import clld
from clld import interfaces
from clld import RESOURCES
from clld.web.util.htmllib import HTML, literal
from clld.web.util.downloadwidget import DownloadWidget
from clld.db.meta import DBSession
from clld.db.models import common as models
from clld.web.adapters import get_adapter, get_adapters
from clld.lib.coins import ContextObject
from clld.lib import bibtex
from clld.lib import rdf

# appease pyflakes
assert get_adapter
assert get_adapters
assert get_current_request
assert datetime
assert xmlchars
assert groupby

#: dimension of marker images on maps, legends and in datatables.
MARKER_IMG_DIM = '20'


def cc_link(req, license_url, button='regular'):
    license_url = license_url.decode('utf8') if isinstance(license_url, bytes) else license_url
    if license_url == 'https://en.wikipedia.org/wiki/Public_domain':
        license_url = 'https://creativecommons.org/publicdomain/zero/1.0/'  # pragma: no cover
    license_url = urlparse(license_url)
    if license_url.hostname != 'creativecommons.org':
        return

    comps = license_url.path.split('/')
    if len(comps) < 3:
        return  # pragma: no cover

    known = {
        'zero': 'Public Domain',
        'by': 'Creative Commons Attribution License',
        'by-nc': 'Creative Commons Attribution-NonCommercial License',
        'by-nc-nd': 'Creative Commons Attribution-NonCommercial-NoDerivatives License',
        'by-nc-sa': 'Creative Commons Attribution-NonCommercial-ShareAlike License',
        'by-nd': 'Creative Commons Attribution-NoDerivatives License',
        'by-sa': 'Creative Commons Attribution-ShareAlike License'}
    if comps[2] not in known:
        return

    icon = 'cc-' + comps[2] + ('-small' if button == 'small' else '') + '.png'
    img_attrs = dict(
        alt=known[comps[2]],
        src=req.static_url('clld:web/static/images/' + icon))
    height, width = (15, 80) if button == 'small' else (30, 86)
    img_attrs.update(height=height, width=width)
    return HTML.a(HTML.img(**img_attrs), href=license_url.geturl(), rel='license')


def maybe_license_link(req, license, **kw):
    cc_link_ = cc_link(req, license, button=kw.pop('button', 'regular'))
    if cc_link_:
        return cc_link_
    license_url = urlparse(license)
    if license_url.hostname:
        return external_link(license_url.geturl(), **kw)
    return license


def get_valueset(req, ctx):
    param = req.params.get('parameter')
    if param is None:  # pragma: no cover
        return

    try:
        param = int(param)
    except ValueError:  # pragma: no cover
        pass

    query = DBSession.query(models.ValueSet)\
        .filter(models.ValueSet.language_pk == ctx.pk)

    if isinstance(param, int):
        query = query.filter(models.ValueSet.parameter_pk == param)
    else:  # pragma: no cover
        query = query.join(models.Parameter).filter(models.Parameter.id == param)

    return query.first()


def get_url_template(req, route, relative=True, variable_map=None):
    variable_map = variable_map or {}
    if isinstance(route, str):
        route = req.registry.getUtility(IRoutesMapper).get_route(route)
    if route:
        res = '' if relative else req.application_url
        param_pattern = re.compile(r'\{(?P<name>[a-z]+)(:[^\}]+)?\}')
        return res + param_pattern.sub(
            lambda m: '{%s}' % variable_map.get(m.group('name'), m.group('name')),
            route.pattern)


def rdf_namespace_attrs():
    return '\n'.join('xmlns:%s="%s"' % item for item in rdf.NAMESPACES.items())


def urlescape(string):
    return quote(string, safe='')


def dumps(obj):
    return JS.sub(pyramid_render('json', obj))


def data_uri(filename, mimetype):
    with open(filename, mode='rb') as fp:
        content = fp.read()
    return 'data:%s;base64,%s' % (mimetype, b64encode(content).decode())


class JS(object):

    """Markup JavaScript object names for proper serialization to pseudo-JSON.

    We use JSON serialization to convert python objects to javascript objects, e.g. for
    options. Sometimes these options take javascript names (of functions, etc.). To make
    sure such names can be serialized properly, i.e. without being quoted as literal
    strings, they can be wrapped into a JS object.
    """

    delimiter = '|'
    pattern = re.compile(r'\"\{0}(?P<name>[^\{0}]+)\{0}\"'.format(delimiter))

    def __init__(self, name):
        self.name = name

    def __call__(self, *args):
        return '%s(%s)' % (self.name, ', '.join(
            arg.name if isinstance(arg, JS) else dumps(arg) for arg in args))

    def __getattr__(self, attr):
        return JS('%s.%s' % (self.name, attr))

    def __json__(self, request=None):
        return '{0}{1}{0}'.format(self.delimiter, self.name)

    @classmethod
    def sub(self, string):
        return self.pattern.sub(lambda m: m.group('name'), string)


JSFeed = JS('CLLD.Feed')
JSMap = JS('CLLD.map')
JSModal = JS('CLLD.Modal')
JSDataTable = JS('CLLD.DataTable')


class JSNamespace(object):

    """Shortcut to create JS objects within a common namespace."""

    def __init__(self, prefix):
        self.prefix = prefix

    def __getattr__(self, name):
        return JS(self.prefix + '.' + name)


JS_CLLD = JSNamespace('CLLD')


def text_citation(req, ctx):
    citation = get_adapter(interfaces.IRepresentation, ctx, req, ext='md.txt')
    return citation.render(ctx, req)


def get_downloads(req):
    for k, items in groupby(
        sorted(
            list(req.registry.getUtilitiesFor(interfaces.IDownload)), key=lambda t: t[0]),
        lambda t: t[1].model.__name__ + 's',
    ):
        yield k, sorted([i[1] for i in items if i[1].size], key=lambda d: d.ext)


def get_rdf_dumps(req, model):
    rdf_exts = [n.extension for n in rdf.FORMATS.values()]
    for name, dl in req.registry.getUtilitiesFor(interfaces.IDownload):
        if dl.model == model and dl.ext in rdf_exts:
            yield dl


def coins(req, obj, label=''):
    if not isinstance(obj, bibtex.Record):
        adapter = get_adapter(interfaces.IMetadata, obj, req, ext='md.bib')
        if not adapter:
            return label
        obj = adapter.rec(obj, req)
    co = ContextObject.from_bibtex('clld', obj)
    return HTML.span(label, **co.span_attrs())


def format_gbs_identifier(source):
    return source.gbs_identifier.replace(':', '-') if source.gbs_identifier else source.pk


def format_coordinates(obj, no_seconds=True, wgs_link=True):
    """Format WGS84 coordinates as HTML.

    .. seealso:: https://en.wikipedia.org/wiki/ISO_6709#Order.2C_sign.2C_and_units
    """
    if not isinstance(obj.latitude, float) or not isinstance(obj.longitude, float):
        return ''
    return HTML.div(
        HTML.table(
            HTML.tr(
                HTML.td(
                    'Coordinates ',
                    external_link(
                        'https://en.wikipedia.org/wiki/World_Geodetic_System_1984',
                        label="WGS84") if wgs_link else ''),
                HTML.td(
                    HTML.span('%s, %s' % (
                        degminsec(obj.latitude, 'NS', no_seconds=no_seconds),
                        degminsec(obj.longitude, 'EW', no_seconds=no_seconds))),
                    HTML.br(),
                    HTML.span(
                        '{0.latitude:.2f}, {0.longitude:.2f}'.format(obj),
                        class_='geo'))),
            class_="table table-condensed"))


def format_license_icon_url(req):
    if 'license_icon' in req.dataset.jsondata:
        url = req.dataset.jsondata['license_icon']
        if not url.startswith('http'):
            url = req.static_url('clld:web/static/images/' + url)
        return url


def format_frequency(req, obj, marker=None, height=MARKER_IMG_DIM, width=MARKER_IMG_DIM):
    if not obj.frequency or obj.frequency == 100:
        return ''
    res = 'Frequency: %s%%' % round(obj.frequency, 1)
    marker = marker or req.registry.queryUtility(interfaces.IFrequencyMarker)
    if marker:
        url = marker(obj, req)
        if url:
            return marker_img(url, height=height, width=width, alt=res, title=res)
    return res


def map_marker_url(req, obj, marker=None):
    marker = marker or req.registry.getUtility(interfaces.IMapMarker)
    return marker(obj, req)


def map_marker_img(req, obj, marker=None, height=MARKER_IMG_DIM, width=MARKER_IMG_DIM):
    url = map_marker_url(req, obj, marker=marker)
    return marker_img(url, height=height, width=width) if url else ''


def link(req, obj, **kw):
    get_link_attrs = req.registry.queryUtility(interfaces.ILinkAttrs)
    if get_link_attrs:
        kw = get_link_attrs(req, obj, **kw)

    if 'class_' in kw:
        kw['class'] = kw['class_']
        del kw['class_']

    rsc = None
    rsc_name = kw.pop('rsc', None)
    for _rsc in RESOURCES:
        if _rsc.interface.providedBy(obj) or _rsc.name == rsc_name:
            rsc = _rsc
            break
    assert rsc
    href = kw.pop('href', req.resource_url(obj, rsc=rsc, **kw.pop('url_kw', {})))
    kw['class'] = ' '.join(
        filter(None, kw.get('class', '').split() + [rsc.interface.__name__[1:]]))
    label = kw.pop('label', str(obj))
    kw.setdefault('title', label)
    return HTML.a(label, href=href, **kw)


def external_link(url, label=None, inverted=False, **kw):
    kw.setdefault('title', label or url)
    kw.setdefault('href', url)
    return HTML.a(icon('share', inverted=inverted), ' ', label or url, **kw)


def maybe_external_link(text, **kw):
    url = urlparse(text)
    if url.hostname and url.scheme in ['http', 'https']:
        return external_link(text, **kw)
    return text


def link_to_map(language):
    return HTML.a(
        icon('icon-globe'),
        title='show %s on map' % language.name,
        href="#map",
        onclick=JS_CLLD.mapShowInfoWindow(None, language.id))


def gbs_link(source, pages=None):
    """Format Google-Books information for source as HTML."""
    if not source or not source.google_book_search_id or not source.jsondata.get('gbs'):
        return ''
    if source.jsondata['gbs']['accessInfo']['viewability'] in ['NO_PAGES']:
        return ''
    pg = 'PP1'
    if pages:
        match = re.search('(?P<startpage>[0-9]+)', pages)
        if match:
            pg = 'PA' + match.group('startpage')
    return HTML.a(
        HTML.img(src="https://www.google.com/intl/en/googlebooks/images/"
                 "gbs_preview_button1.gif"),
        href="https://books.google.com/books?id=%s&lpg=PP1&pg=%s" % (
            source.google_book_search_id, pg)
    )


def button(*content, **attrs):
    tag = attrs.pop('tag', HTML.a if 'href' in attrs else HTML.button)
    attrs.setdefault('type', 'button')
    class_ = attrs['class'] if 'class' in attrs else attrs.pop('class_', '').split()
    if 'btn' not in class_:
        class_.append('btn')
    attrs['class'] = ' '.join(class_)
    return tag(*content, **attrs)


def cite_button(req, ctx):
    return button(
        'cite',
        id="cite-button-%s" % ctx.id,
        onclick=JSModal.show(ctx.name, req.resource_url(ctx, ext='md.html')))


# regex to match standard abbreviations in gloss units:
# We look for sequences of uppercase letters which are not followed by a lowercase letter.
GLOSS_ABBR_PATTERN = re.compile(
    '(?P<personprefix>1|2|3)?(?P<abbr>[A-Z]+)(?P<personsuffix>1|2|3)?(?=([^a-z]|$))')
ALT_TRANSLATION_LANGUAGE_PATTERN = re.compile(r'((?P<language>[a-zA-Z\s]+):)')


#
# TODO: enumerate exceptions: 1SG, 2SG, 3SG, ?PL, ?DU
#
def rendered_sentence(sentence, abbrs=None, fmt='long'):
    """Format a sentence as HTML."""
    if sentence.xhtml:
        return HTML.div(
            HTML.div(Markup(sentence.xhtml), class_='body'), class_="sentence")

    if abbrs is None:
        q = DBSession.query(models.GlossAbbreviation).filter(
            or_(models.GlossAbbreviation.language_pk == sentence.language_pk,
                models.GlossAbbreviation.language_pk == None)
        )
        abbrs = dict((g.id, g.name) for g in q)

    def gloss_with_tooltip(gloss):
        person_map = {
            '1': 'first person',
            '2': 'second person',
            '3': 'third person',
        }

        res, end = [], 0
        for match in GLOSS_ABBR_PATTERN.finditer(gloss):
            if match.start() > end:
                res.append(gloss[end:match.start()])

            abbr = match.group('abbr')
            if abbr in abbrs:
                explanation = abbrs[abbr]
                if match.group('personprefix'):
                    explanation = '%s %s' % (
                        person_map[match.group('personprefix')], explanation)

                if match.group('personsuffix'):
                    explanation = '%s %s' % (
                        explanation, person_map[match.group('personsuffix')])

                res.append(HTML.span(
                    HTML.span(gloss[match.start():match.end()].lower(), class_='sc'),
                    **{'data-hint': explanation, 'class': 'hint--bottom'}))
            else:
                res.append(
                    (match.group('personprefix') or '') +  # noqa: W504
                    abbr +  # noqa: W504
                    (match.group('personsuffix') or ''))

            end = match.end()

        res.append(gloss[end:])
        return filter(None, res)

    def alt_translation(sentence):
        res = ''
        if sentence.jsondata.get('alt_translation'):
            text = sentence.jsondata['alt_translation']
            name = ''
            if ALT_TRANSLATION_LANGUAGE_PATTERN.match(text):
                name, text = [t.strip() for t in text.split(':', 1)]
                name = HTML.span(name + ': ')
            res = HTML.div(name, HTML.span(text, class_='translation'))
        return res

    units = []
    if sentence.analyzed and sentence.gloss:
        analyzed = sentence.analyzed
        glossed = sentence.gloss
        for morpheme, gloss in zip(analyzed.split('\t'), glossed.split('\t')):
            units.append(HTML.div(
                HTML.div(morpheme, class_='morpheme'),
                HTML.div(*gloss_with_tooltip(gloss), **{'class': 'gloss'}),
                class_='gloss-unit'))

    return HTML.div(
        HTML.div(
            HTML.div(
                HTML.div(sentence.original_script, class_='original-script')
                if sentence.original_script else '',
                HTML.div(literal(sentence.markup_text or sentence.name),
                         class_='object-language'),
                HTML.div(*units, **{'class': 'gloss-box'}) if units else '',
                HTML.div(sentence.description, class_='translation')
                if sentence.description else '',
                alt_translation(sentence),
                class_='body',
            ),
            class_="sentence",
        ),
        class_="sentence-wrapper",
    )


def icon(class_, inverted=False, **kw):
    if not class_.startswith('icon-'):
        class_ = 'icon-' + class_
    if inverted:
        class_ = '%s icon-white' % class_
    return HTML.i(class_=class_, **kw)


def contactmail(req, ctx=None, title='contact maintainer'):
    """Format the contact address for a dataset as mailto: link."""
    params = {}
    for name in ['subject', 'body']:
        params[name] = pyramid_render(
            'contactmail_{0}.mako'.format(name), {'req': req, 'ctx': ctx}, request=req)\
            .strip()\
            .encode('utf8')
        query = urlencode(params).replace('+', '%20')
    href = 'mailto:{0}?{1}'.format(req.contact_email_address, query)
    return button(icon('bell'), title=title, href=href, class_='btn-warning btn-mini')


def iter_join(items, sep):
    for i, item in enumerate(items):
        if i > 0:
            yield sep
        yield item


def newline2br(text):
    """Replace newlines in text with HTML br tags."""
    if not text:
        return ''
    return '\n'.join(iter_join([literal(line) for line in text.split('\n')], HTML.br()))


def text2html(text, mode='br', sep='\n\n'):
    """Turn plain text into simple HTML."""
    if mode == 'p':
        return HTML.div(*[HTML.p(literal(newline2br(line))) for line in text.split(sep)])
    return HTML.p(literal(newline2br(text)))


def linked_contributors(req, contribution):
    chunks = list(iter_join([link(req, c) for c in contribution.primary_contributors], ' and '))
    sec = list(iter_join([link(req, c) for c in contribution.secondary_contributors], ' and '))
    if sec:
        chunks.append(' with ')
        chunks.extend(sec)
    return HTML.span(*chunks)


def linked_references(req, obj):
    chunks = []
    for i, ref in enumerate(getattr(obj, 'references', [])):
        if ref.source:
            gbs = gbs_link(ref.source, pages=ref.description)
            if i > 0:
                chunks.append('; ')
            chunks.append(HTML.span(
                link(req, ref.source),
                HTML.span(
                    ': %s' % ref.description if ref.description else '',
                    class_='pages'),
                ' ' if gbs else '',
                gbs,
                class_='citation',
            ))
    if chunks:
        return HTML.span(*chunks)
    return ''


def language_identifier(req, obj, **kw):
    if not obj:
        return ''

    label = obj.name or obj.id

    for type_ in models.IdentifierType:
        if obj.type == type_.value:
            label = external_link(type_.args[0].format(obj), label=label, **kw)
            break

    return HTML.span(label, class_='language_identifier %s' % obj.type)


def get_referents(source, exclude=None):
    """Retrieve objects referencing source.

    :return: dict storing lists of objects referring to source keyed by type.
    """
    res = {}
    for obj_cls, ref_cls in [
        (models.Language, models.LanguageSource),
        (models.ValueSet, models.ValueSetReference),
        (models.Sentence, models.SentenceReference),
        (models.Contribution, models.ContributionReference),
    ]:
        if obj_cls.__name__.lower() in (exclude or []):
            continue
        q = DBSession.query(obj_cls).join(ref_cls).filter(ref_cls.source_pk == source.pk)
        if obj_cls == models.ValueSet:
            q = q.options(
                joinedload(models.ValueSet.parameter),
                joinedload(models.ValueSet.language))
        res[obj_cls.__name__.lower()] = q.all()
    return res


def alt_representations(req, rsc, doc_position='right', exclude=None):
    """Represent available adapters for rsc as dropdown menu."""
    kw = dict(doc_position=doc_position)
    if exclude:
        kw['exclude'] = exclude
    route, route_kw = req._route(rsc, None, ext='%s')
    dlw = DownloadWidget(
        req,
        rsc,
        rsc,
        JS_CLLD.route_url(route, route_kw),
        interfaces.IRepresentation,
        **kw)
    return dlw.render()


def partitioned(items, n=3):
    """Partition items into n buckets."""
    max_items_per_bucket, rem = divmod(len(items), n)
    if rem:
        max_items_per_bucket += 1
    bucket = []

    for item in items:
        if len(bucket) >= max_items_per_bucket:
            yield bucket
            bucket = []
        bucket.append(item)

    yield bucket


def marker_img(src, **kw):
    kw.setdefault('height', MARKER_IMG_DIM)
    kw.setdefault('width', MARKER_IMG_DIM)
    return HTML.img(src=src, **kw)


def glottolog_url(glottocode):
    return models.Identifier(name=glottocode, type='glottolog').url()


def collapsed(id_, content, button_content=None):
    return HTML.div(
        HTML.p(HTML.a(
            button_content or icon('plus-sign'),
            **{'class': 'btn', 'data-toggle': 'collapse', 'data-target': '#%s' % id_})),
        HTML.div(content, id=id_, class_='collapse'))


def static_path(*comps):
    return os.path.abspath(
        os.path.join(os.path.dirname(clld.__file__), 'web', 'static', *comps))


def charis_font_spec_css():
    """Font spec for using CharisSIL with Pisa (xhtml2pdf)."""
    return """
    @font-face {{
        font-family: 'charissil';
        src: url('{0}/CharisSIL-Regular.ttf');
    }}
    @font-face {{
        font-family: 'charissil';
        font-style: italic;
        src: url('{0}/CharisSIL-Italic.ttf');
    }}
    @font-face {{
        font-family: 'charissil';
        font-weight: bold;
        src: url('{0}/CharisSIL-Bold.ttf');
    }}
    @font-face {{
        font-family: 'charissil';
        font-weight: bold;
        font-style: italic;
        src: url('{0}/CharisSIL-BoldItalic.ttf');
    }}
""".format(static_path('fonts'))


def get_resource_type(obj):
    """We rely on resource objects to implement just one interface.

    :param obj:
    :return: resource type as string
    """
    for interface in providedBy(obj):
        return interface.__name__[1:].lower()


def localize_url(req, locale, default_locale='en'):
    """Add localization query parameter to current url.
    If locale = default_locale it removes the query parameter.

    :param req: current request
    :param locale: language code (ISO 639-1:2002)
    :param default_locale: default language code (ISO 639-1:2002)
    :return: URL as string
    """
    assert locale, 'Missing locale'
    k_str = '__locale__'
    p_url = urlsplit(req.url)
    params = {
        k: v for k, v in req.query_params.items()
        if not (k == k_str and locale.lower() == default_locale.lower())}
    if locale.lower() != default_locale.lower():
        params[k_str] = locale.lower()
    return urlunsplit(
        SplitResult(p_url.scheme, p_url.netloc, p_url.path,
                    urlencode(params, doseq=True),
                    p_url.fragment))



"""
Functionality to serialize clld objects as GeoJSON.

.. seealso:: http://geojson.org/
"""
import json
import itertools

from zope.interface import implementer
from pyramid.renderers import render as pyramid_render
from sqlalchemy.orm import joinedload
from clldutils.misc import nfilter

from clld.web.adapters.base import Renderable
from clld.web.icon import Icon
from clld import interfaces
from clld.db.meta import DBSession
from clld.db.models.common import ValueSet, Value, Language


_PACIFIC_CENTERED = False


def pacific_centered():
    global _PACIFIC_CENTERED
    _PACIFIC_CENTERED = True


def _flatten(d, parent_key=''):
    items = []

    if isinstance(d, list):
        for i, v in enumerate(d):
            items.extend(_flatten(v, '_'.join(filter(None, [parent_key, str(i)]))))
    elif isinstance(d, dict):
        for k, v in d.items():
            new_key = parent_key + '_' + k if parent_key else k
            items.extend(_flatten(v, new_key))
    else:
        items.append((parent_key, d))
    return items


def flatten(d):
    """tilemill can only handle flat properties objects.

    >>> sorted(flatten({'a': {'b': [1, {'c': 4}, 3]}}).keys())
    ['a_b_0', 'a_b_1_c', 'a_b_2']
    """
    return dict(_flatten(d))


def get_lonlat(obj):
    """
    Return coordinates of an object taking into account whether maps are pacific centered.

    Pacific centered means
    the world should be divided between Icelandic (westernmost language) and
    Tupi (easternmost language), i.e. between -17 and -36.

    We chose -26 as divider because that puts cape verde to the west of africa.

    .. seealso: https://github.com/Leaflet/Leaflet/issues/1360
    """
    if hasattr(obj, 'latitude') and hasattr(obj, 'longitude'):
        longitude, latitude = obj.longitude, obj.latitude
    elif isinstance(obj, (tuple, list)) and len(obj) == 2:
        longitude, latitude = obj
    else:
        return
    if longitude is None or latitude is None:
        return
    if _PACIFIC_CENTERED and longitude <= -26:
        longitude += 360
    return longitude, latitude


def get_feature(obj, lonlat=None, **properties):
    res = {
        'type': 'Feature',
        'properties': properties,
        'geometry': {'type': 'Point', 'coordinates': lonlat or get_lonlat(obj)}}
    if hasattr(obj, 'id'):
        res['id'] = obj.id
    if hasattr(obj, 'name'):
        res['properties'].setdefault('name', obj.name)
    return res


@implementer(interfaces.IRepresentation)
class GeoJson(Renderable):

    """Base class for adapters which render geojson feature collections.

    The geojson we serve to leaflet must fulfill the following requirements:
    - a layer member in the featurecollection properties.
    - an icon member in the feature properties.
    - a language.id member in feature properties.
    """

    name = "GeoJSON"
    extension = 'geojson'
    mimetype = 'application/geojson'
    send_mimetype = 'application/json'  # application/vnd.geo+json

    def _featurecollection_properties(self, ctx, req):
        """Get properties object for the FeatureCollection.

        We return the layer index passed in the request, to make sure the features are
        added to the correct layer group.
        """
        res = {'layer': req.params.get('layer', '')}
        res.update(self.featurecollection_properties(ctx, req))
        return res

    def featurecollection_properties(self, ctx, req):
        """override to add properties."""
        return {}

    def feature_iterator(self, ctx, req):
        """Yield objects which will be represented as GeoJSON features.

        Since the only objects we have geographic info about are Languages, we try to
        detect (collections of) Languages associated with ctx.

        :return: generator object.
        """
        if interfaces.IDataTable.providedBy(ctx) and ctx.model == Language:
            return ctx.get_query(limit=5000)
        if hasattr(ctx, 'languages'):
            return ctx.languages
        if interfaces.ILanguage.providedBy(ctx):
            return [ctx]
        return []

    def feature_properties(self, ctx, req, feature):
        """override to add properties."""
        return {}

    def get_language(self, ctx, req, feature):
        """override to fetch language object from non-default location."""
        return feature

    def get_features(self, ctx, req):
        map_marker = req.registry.getUtility(interfaces.IMapMarker)

        for feature in self.feature_iterator(ctx, req):
            language = self.get_language(ctx, req, feature)
            lonlat = get_lonlat(language)
            if lonlat:
                properties = self.feature_properties(ctx, req, feature) or {}
                properties.setdefault('icon', map_marker(feature, req))
                properties.setdefault('language', language)
                yield get_feature(language, lonlat=lonlat, **properties)

    def render(self, ctx, req, dump=True):
        res = {
            'type': 'FeatureCollection',
            'properties': self._featurecollection_properties(ctx, req),
            'features': list(self.get_features(ctx, req))}
        return pyramid_render('json', res, request=req) if dump else res


class GeoJsonParameter(GeoJson):

    """Render a parameter's values as geojson feature collection."""

    def featurecollection_properties(self, ctx, req):
        marker = req.registry.getUtility(interfaces.IMapMarker)
        return {
            'name': getattr(ctx, 'name', 'Values'),
            'domain': [
                {'icon': marker(de, req), 'id': de.id, 'name': de.name}
                for de in getattr(ctx, 'domain', [])]}

    def get_query(self, ctx, req):
        return DBSession.query(ValueSet)\
            .join(Value)\
            .filter(ValueSet.parameter_pk == ctx.pk)\
            .options(joinedload(ValueSet.values), joinedload(ValueSet.language))

    def feature_iterator(self, ctx, req):
        de = req.params.get('domainelement')
        if de:
            return [vs for vs in ctx.valuesets
                    if vs.values and vs.values[0].domainelement.id == de]
        return self.get_query(ctx, req)

    def get_language(self, ctx, req, valueset):
        return valueset.language

    def feature_properties(self, ctx, req, valueset):
        return {
            'values': list(valueset.values),
            'label': ', '.join(nfilter(v.name for v in valueset.values))
            or self.get_language(ctx, req, valueset).name}


class GeoJsonParameterMultipleValueSets(GeoJsonParameter):

    """GeoJSON adapter for parameters with multiple valuesets per language."""

    def feature_iterator(self, ctx, req):
        q = self.get_query(ctx, req)
        return itertools.groupby(q.order_by(ValueSet.language_pk), lambda vs: vs.language)

    def get_language(self, ctx, req, pair):
        return pair[0]

    def feature_properties(self, ctx, req, pair):
        return {}


class GeoJsonCombinationDomainElement(GeoJson):

    """GeoJSON adapter for a domain element of a combination of parameters."""

    def feature_properties(self, ctx, req, language):
        icon = Icon.from_req(ctx, req) or ctx.icon
        return {
            'icon': icon.url(req) if icon else '',
            'zindex': 1000 - len(ctx.languages)}


class GeoJsonParameterFlatProperties(GeoJsonParameter):

    """GeoJSON with flattened feature properties for importing in tilemill.

    .. seealso: https://github.com/mapbox/tilemill/issues/2294
    """

    name = "GeoJSON for tilemill"
    extension = 'flat.geojson'
    mimetype = 'application/flat+geojson'

    def render(self, ctx, req, dump=True):
        res = json.loads(GeoJson.render(self, ctx, req, dump=True))
        for f in res['features']:
            f['properties'] = flatten(f['properties'])
        return pyramid_render('json', res, request=req) if dump else res


@implementer(interfaces.IIndex)
class GeoJsonLanguages(GeoJson):
    """Render a collection of languages as geojson feature collection."""



import gzip
from contextlib import closing
from xml.etree import cElementTree as et

from clld.db.models.common import Language, Source
from clld.web.adapters.download import *


def test_download_dir():
    assert download_dir('clld')


def test_Download(mocker, env, tmp_path):
    dl = Download(Source, 'clld', ext='x')
    assert dl.asset_spec(mocker.Mock()).startswith('clld:')

    class TestDownload(Download):
        def asset_spec(self, req):
            return str(tmp_path / 'x')

    dl = TestDownload(Source, 'clld', ext='bib')
    abspath = dl.abspath(env['request'])
    assert not abspath.exists()
    dl.create(env['request'], verbose=False)
    dl.size(env['request'])
    dl.label(env['request'])
    assert abspath.exists()
    abspath.unlink()

    dl = TestDownload(Source, 'clld', ext='rdf')
    dl.create(env['request'], verbose=False)
    dl.abspath(env['request']).unlink()


def test_Download_url(mocker, env):
    mocker.patch.multiple(
        'clld.web.adapters.download',
        download_asset_spec=mocker.Mock(
            return_value='clld:web/static/images/favicon.ico'))
    dl = Download(Source, 'clld', ext='bib')
    assert dl.url(env['request'])


def test_Download2(env, tmp_path):
    out = tmp_path / 'dl'
    dl = CsvDump(Language, 'clld')
    dl.create(env['request'], verbose=False, outfile=out)
    assert out.exists()
    dl.create(env['request'], filename=out, verbose=False, outfile=out)
    dl = N3Dump(Language, 'clld')
    assert dl.abspath(env['request']).name == 'dataset-language.n3.gz'
    dl.create(env['request'], verbose=False, outfile=out)
    dl = RdfXmlDump(Language, 'clld')
    dl.create(env['request'], verbose=False, outfile=out)

    with closing(gzip.open(out.as_posix(), 'rb')) as fp:
        assert et.fromstring(fp.read())



from clld.db.models import common
from clld.web.util.multiselect import *


def test_MultiSelect(env):
    ms = MultiSelect(env['request'], common.Language, 'x', url='/')
    ms.render()
    ms.render(selected=[common.Language.first()])
    ms.format_result(common.Language(id='x'))


def test_CombinationMultiSelect(env):
    ms = CombinationMultiSelect(
        env['request'], combination=common.Combination(common.Parameter.first()))
    ms.render()



"""Default DataTable for UnitValue objects."""
from sqlalchemy.orm import joinedload

from clld.db.models import common
from clld.web.datatables.base import DataTable, LinkCol


class UnitValueNameCol(LinkCol):

    """render the label for the UnitValue."""

    def order(self):
        return common.UnitDomainElement.id \
            if self.dt.unitparameter and self.dt.unitparameter.domain \
            else [common.UnitValue.name, common.UnitValue.id]

    def search(self, qs):
        if self.dt.unitparameter and self.dt.unitparameter.domain:
            return common.UnitDomainElement.name.contains(qs)
        return common.UnitValue.name.contains(qs)


class Unitvalues(DataTable):

    """Default DataTable for UnitValue objects."""

    __constraints__ = [common.UnitParameter, common.Contribution, common.Unit]

    def base_query(self, query):
        query = query\
            .join(common.Unit)\
            .outerjoin(common.UnitDomainElement)\
            .options(joinedload(common.UnitValue.unit))

        if self.unit:
            return query.filter(common.UnitValue.unit_pk == self.unit.pk)

        if self.unitparameter:
            return query.filter(
                common.UnitValue.unitparameter_pk == self.unitparameter.pk)

        if self.contribution:
            return query.filter(common.UnitValue.contribution_pk == self.contribution.pk)

        return query

    def col_defs(self):
        name_col = UnitValueNameCol(self, 'value')
        if self.unitparameter and self.unitparameter.domain:
            name_col.choices = sorted([de.name for de in self.unitparameter.domain])
        return [
            name_col,
            LinkCol(self, 'unit', get_obj=lambda i: i.unit, model_col=common.Unit.name),
        ]

    def toolbar(self):
        return ''



from datetime import date

import pytest

from testutils import XmlResponse
from clld.web.views.olac import ResumptionToken, olac, OlacConfig


class OaiPmhResponse(XmlResponse):
    ns = 'http://www.openarchives.org/OAI/2.0/'

    @property
    def error(self):
        e = self.findall('error')
        if e:
            return e[0].get('code')


def test_ResumptionToken():
    assert str(ResumptionToken(from_=date.today(), until=date.today()))


@pytest.mark.parametrize(
    "params,test,expected",
    [
        ({}, lambda res: res.error, 'badVerb'),
        ({'verb': 'ListSets'}, lambda res: res.error, 'noSetHierarchy'),
        (dict(verb='Identify', other='arg'), lambda res: res.error, 'badArgument'),
        (dict(verb='Identify'), lambda res: len(res.findall('Identify')), 1),
        (
            dict(verb='ListMetadataFormats'),
            lambda res: res.findone('metadataPrefix').text,
            'olac'),
        (
            dict(verb='ListMetadataFormats', other='x'),
            lambda res: res.error,
            'badArgument'),
        (
            dict(verb='ListIdentifiers', metadataPrefix='olac'),
            lambda res: len(res.findall('header')),
            100),
        (
            dict(verb='Identify'),
            lambda res: res.findone(
                '{http://www.openarchives.org/OAI/2.0/oai-identifier}sampleIdentifier'
            ).text,
            'oai:clld:language'),
        (
            dict(verb='GetRecord', metadataPrefix='olac', identifier='oai:clld:language'),
            lambda res: res.findone('identifier').text,
            'oai:clld:language'),
        (
            dict(verb='GetRecord', metadataPrefix='olac'),
            lambda res: res.error,
            'badArgument'),
        (
            dict(verb='GetRecord', metadataPrefix='ol', identifier='oai:clld:language'),
            lambda res: res.error,
            'cannotDisseminateFormat'),
        (
            dict(verb='GetRecord', metadataPrefix='olac', identifier='oai:clld:123'),
            lambda res: res.error,
            'idDoesNotExist'),
        (
            dict(verb='ListIdentifiers', resumptionToken='tr', metadataPrefix='olac'),
            lambda res: res.error,
            'badArgument'),
        (
            dict(verb='ListIdentifiers', resumptionToken='tr', o='x'),
            lambda res: res.error,
            'badArgument'),
        (
            dict(verb='ListIdentifiers'),
            lambda res: res.error,
            'badArgument'),
        (
            dict(verb='ListIdentifiers', metadataPrefix='olac', set='x'),
            lambda res: res.error,
            'noSetHierarchy'),
        (
            dict(verb='ListIdentifiers', resumptionToken='tr'),
            lambda res: res.error,
            'badResumptionToken'),
        (
            dict(verb='ListIdentifiers', resumptionToken='0f2000-01-01u2222-01-01'),
            lambda res: res.error,
            None),
        (
            dict(verb='ListIdentifiers', resumptionToken='100'),
            lambda res: res.error,
            None),
        (
            dict(verb='ListIdentifiers', resumptionToken='200'),
            lambda res: res.error,
            'noRecordsMatch'),
        (
            dict(verb='ListIdentifiers', resumptionToken='100f2000-01-01u2000-01-01'),
            lambda res: res.error,
            'noRecordsMatch'),
    ])
def test_olac(request_factory, params, test, expected):
    with request_factory(params=params) as req:
        assert test(OaiPmhResponse(olac(req))) == expected


def test_olac_config():
    OlacConfig()



from sqlalchemy import Column, Integer, String, Unicode, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces
from clld.lib import bibtex
from clld.lib import coins
from clld.web.util.htmllib import HTML

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin,
    Language, LanguageSource)

__all__ = ('Source', 'HasSourceMixin', 'HasSourceNotNullMixin')


class Source_data(Base, DataMixin):
    pass


class Source_files(Base, FilesMixin):
    pass


@implementer(interfaces.ISource)
class Source(Base,
             PolymorphicBaseMixin,
             IdNameDescriptionMixin,
             HasDataMixin,
             HasFilesMixin):

    """A bibliographic record, cited as source for some statement."""

    glottolog_id = Column(String)
    google_book_search_id = Column(String)

    #
    # BibTeX fields:
    #
    bibtex_type = Column(bibtex.EntryType.db_type())
    author = Column(Unicode)
    year = Column(Unicode)
    title = Column(Unicode)
    type = Column(Unicode)
    booktitle = Column(Unicode)
    editor = Column(Unicode)
    pages = Column(Unicode)
    edition = Column(Unicode)
    journal = Column(Unicode)
    school = Column(Unicode)
    address = Column(Unicode)
    url = Column(Unicode)
    note = Column(Unicode)
    number = Column(Unicode)
    series = Column(Unicode)
    volume = Column(Unicode)
    publisher = Column(Unicode)
    organization = Column(Unicode)
    chapter = Column(Unicode)
    howpublished = Column(Unicode)

    # typed information we might want to use for searching or sorting:
    year_int = Column(Integer)
    startpage_int = Column(Integer)
    pages_int = Column(Integer)

    languages = relationship(
        Language, backref='sources', secondary=LanguageSource.__table__)

    @property
    def gbs_identifier(self):
        if not self.jsondata.get('gbs'):
            return
        if not self.jsondata['gbs']['volumeInfo'].get('industryIdentifiers'):
            return
        id_ = None
        for identifier in self.jsondata['gbs']['volumeInfo']['industryIdentifiers']:
            # prefer ISBN_13 over ISBN_10 over anything else
            if identifier['type'] == 'ISBN_13':
                id_ = 'ISBN:' + identifier['identifier']
            if identifier['type'] == 'ISBN_10' and not id_:
                id_ = 'ISBN:' + identifier['identifier']
        if not id_:
            # grab the last one in the list (most probably the only one!)
            id_ = identifier['identifier']
        return id_

    def __bibtex__(self):
        return {}

    def bibtex(self, exclude={'gbs', 'glottolog_ref_id'}):
        kw = {k: v for k, v in self.jsondata.items() if k not in exclude}
        kw.update(self.__bibtex__())
        return bibtex.Record.from_object(self, **kw)

    def coins(self, req):
        return HTML.span(
            ' ',
            **coins.ContextObject.from_bibtex(
                req.dataset.name, self.bibtex()).span_attrs()
        )


#
# Several objects can be linked to sources, i.e. they can have references.
#
class HasSourceMixin(object):
    key = Column(Unicode)  # the citation key, specific (and unique) within a contribution
    description = Column(Unicode)  # e.g. page numbers.

    @declared_attr
    def source_pk(cls):  # pragma: no cover
        return Column(Integer, ForeignKey('source.pk'))

    @declared_attr
    def source(cls):  # pragma: no cover
        return relationship(Source, backref=cls.__name__.lower() + 's')


class HasSourceNotNullMixin(HasSourceMixin):

    @declared_attr
    def source_pk(cls):
        return Column(Integer, ForeignKey('source.pk'), nullable=False)

    @declared_attr
    def source(cls):
        return relationship(Source, innerjoin=True, backref=cls.__name__.lower() + 's')



from zope.interface import (
    Interface,
    Attribute,
)


# ---------------------------------------------------------------------------
# Interfaces for model classes
# ---------------------------------------------------------------------------
class IDataset(Interface):

    """marker."""


class IFile(Interface):

    """marker."""


class ILanguage(Interface):
    name = Attribute('')
    latitude = Attribute('')
    longitude = Attribute('')


class ISentence(Interface):

    """marker."""


class ISource(Interface):

    """marker."""


class IParameter(Interface):

    """marker."""


class IValueSet(Interface):

    """marker."""


class IValue(Interface):

    """marker."""


class IContribution(Interface):

    """marker."""


class IContributor(Interface):

    """marker."""


class IDomainElement(Interface):

    """marker."""


class IUnitParameter(Interface):

    """marker."""


class IUnit(Interface):

    """marker."""


class IUnitValue(Interface):

    """marker."""


class IUnitDomainElement(Interface):

    """marker."""


class ICombination(Interface):

    """marker."""

    pass


# ---------------------------------------------------------------------------
# Interfaces for pluggable functionality
# ---------------------------------------------------------------------------
class IRepresentation(Interface):
    mimetype = Attribute('')
    charset = Attribute('')
    extension = Attribute('')

    def render(request):
        """ """

    def render_to_response(request):
        """ """


class IIndex(Interface):
    mimetype = Attribute('')
    charset = Attribute('')
    extension = Attribute('')

    def render(request, query, count):
        """ """

    def render_to_response(request, query, count):
        """ """


class IMetadata(Interface):

    """marker."""


class IDownload(Interface):
    name = Attribute('name the download will be registered for')

    def url(request):
        """called from the template for the download page."""

    def create(request):
        """called from the script creating all downloads."""


class IDataTable(Interface):
    model = Attribute('model class of the objects listed in the table')

    def get_query(request, model, **kw):
        """called to retrieve a filtered and sorted sqla query."""


class IMap(Interface):

    """marker."""


class IMenuItems(Interface):

    """marker."""


class IMapMarker(Interface):

    """utility."""

    def __call__(self, ctx, req):
        """Return URL for a marker to use in maps."""


class IIcon(Interface):

    """utility to map icon names to URLs."""

    name = Attribute('name of the icon')

    def url(self, req):
        """return URL for the marker."""


class IFrequencyMarker(Interface):

    """utility."""

    def __call__(self, ctx, req):
        """return URL for an icon to use as marker for frequency."""


class ILinkAttrs(Interface):

    """utility to customize attributes for HTML links to objects."""

    def __call__(self, req, obj, **kw):
        """return dictionary of attributes for link creation."""


class ICtxFactoryQuery(Interface):

    """utility."""

    def __call__(self, model, req):
        """modified query."""


class IOlacConfig(Interface):

    """configuration of an applications OLAC repository."""

    scheme = Attribute('oai-pmh identifier scheme')
    delimiter = Attribute('oai-pmh identifier delimiter')

    def get_earliest_record(self, req):
        """ """

    def get_record(self, req, identifier):
        """ """

    def query_records(self, req, from_=None, until=None):
        """ """

    def format_identifier(self, req, id_):
        """ """

    def parse_identifier(self, req, id_):
        """ """


class IStaticResource(Interface):

    """A resource to be linked from the app template."""

    type = Attribute('js or css')
    asset_spec = Attribute('asset spec to be passed to request.static_url')



from clld.db.models import common

from testutils import handle_dt


def test_Languages(env):
    from clld.web.datatables.language import Languages

    handle_dt(env['request'], Languages, common.Language)



"""Default DataTable for Contributor objects."""
from clld.web.datatables.base import DataTable, Col, LinkCol, ExternalLinkCol
from clld.web.util.htmllib import HTML
from clld.web.util.helpers import link, text2html
from clld.db.models.common import Contributor


class ContributionsCol(Col):

    """Render a list of contributions a Contributor has participated in."""

    __kw__ = {'bSearchable': False, 'bSortable': False}

    def format(self, item):
        return HTML.ul(
            *[HTML.li(link(
                self.dt.req, c.contribution)) for c in item.contribution_assocs])


class AddressCol(Col):
    def format(self, item):
        return text2html(item.address)


class NameCol(LinkCol):
    def order(self):
        return Contributor.id


class UrlCol(ExternalLinkCol):

    """Render a link to a Contributor's homepage."""

    def get_attrs(self, item):
        return {'label': 'homepage'}


class Contributors(DataTable):

    """Default DataTable for Contributor objects."""

    def col_defs(self):
        return [
            NameCol(self, 'name'),
            ContributionsCol(self, 'Contributions'),
            AddressCol(self, 'address'),
            UrlCol(self, 'Homepage'),
        ]



from zope.interface import implementer
from sqlalchemy import (
    Column,
    String,
    Unicode,
    Integer,
    Boolean,
    ForeignKey,
    UniqueConstraint,
)
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property

from clld import interfaces
from clld.db.meta import Base, CustomModelMixin
from clld.db.models import common

#-----------------------------------------------------------------------------
# specialized common mapper classes
#-----------------------------------------------------------------------------
{% if cookiecutter.cldf_module %}
@implementer(interfaces.ILanguage)
class Variety(CustomModelMixin, common.Language):
    pk = Column(Integer, ForeignKey('language.pk'), primary_key=True)
    glottocode = Column(Unicode)
{% endif %}

{% if cookiecutter.cldf_module.lower() == 'wordlist' %}

@implementer(interfaces.IParameter)
class Concept(CustomModelMixin, common.Parameter):
    pk = Column(Integer, ForeignKey('parameter.pk'), primary_key=True)
    concepticon_id = Column(Unicode)

{% elif cookiecutter.cldf_module.lower() == 'structuredataset' %}

@implementer(interfaces.IParameter)
class Feature(CustomModelMixin, common.Parameter):
    pk = Column(Integer, ForeignKey('parameter.pk'), primary_key=True)

{% endif %}


"""Functionality to manage icons for map markers."""
import itertools

from clldutils import svg
from clldutils.color import rgb_as_hex
from zope.interface import implementer

from clld.interfaces import IIcon, IMapMarker, IValue, IValueSet, IDomainElement

SHAPES = [
    "c",  # circle
    "s",  # square
    "t",  # triangle (pyramid)
    "f",  # inverted pyramid
    "d",  # diamond
]

COLORS = [
    '000000',
    '0000dd',
    '0000ff',
    '009900',
    '00ff00',
    '00ffff',
    '4d6cee',
    '66ff33',
    '990099',
    '9999ff',
    '99ffff',
    'a0fb75',
    'aa0000',
    'cb9a34',
    'cccccc',
    'd22257',
    'dd0000',
    'e8e8e8',
    'ed9c07',
    'efe305',
    'f38847',
    'f3ffb0',
    'fe3856',
    'ff0000',
    'ff00ff',
    'ff4400',
    'ff6600',
    'ff66ff',
    'ffcc00',
    'ffff00',
    'ffffcc',
    'ffffff',
]
DEFAULT_ICON = 'cff6600'

# the following colors show enough mutual contrast to be easily distinguised:
PREFERED_COLORS = [
    '0000dd',
    '009900',
    '990099',
    'dd0000',
    'ffff00',
    'ffffff',
    '00ff00',
    '00ffff',
    'cccccc',
    'ff6600',
]
SECONDARY_COLORS = [c for c in COLORS if c not in PREFERED_COLORS]


@implementer(IIcon)
class Icon(object):

    """Default implementation of IIcon: Icons are static image files."""

    def __init__(self, name, opacity=None, select_id=None):
        self.name = name
        self.opacity = opacity
        self.select_id = select_id

    @classmethod
    def from_req(cls,
                 ctx,
                 req,
                 icon_spec_factory=None,
                 icon_map=None):
        from clld.db.models.common import CombinationDomainElement  # Avoid circular imports.

        de, icon, opacity, select_id = None, None, '1.0', None
        icon_map = icon_map or {}

        if IValue.providedBy(ctx):
            de = ctx.domainelement
        elif IValueSet.providedBy(ctx):
            de = ctx.values[0].domainelement
        elif IDomainElement.providedBy(ctx):
            de = ctx
        elif isinstance(ctx, CombinationDomainElement):
            select_id = 'v' + ctx.id
            icon = req.params.get(select_id, ctx.icon.name)

        if de:
            select_id = 'v{}'.format(de.number)
            icon = req.params.get(select_id, de.jsondata['icon'])

        if not icon and icon_spec_factory:
            icon = icon_spec_factory(ctx, req)
            if isinstance(icon, tuple):
                icon, select_id = icon

        if icon:
            opacity = None
            if "'" in icon:
                icon = icon.split("'")[0]
            if len(icon) > 4:
                if len(icon) == 9:
                    opacity = int('0x' + icon[7:], base=16) / 255
                    icon = icon[:7]
                elif len(icon) == 7:
                    pass
                else:
                    icon = icon[:4]
            if len(icon) == 4:
                icon = icon[0] + 2 * icon[1] + 2 * icon[2] + 2 * icon[3]
            return cls(icon_map.get(icon, icon), opacity=opacity, select_id=select_id)

    @property
    def shape(self):
        return self.name[0]

    @property
    def color(self):
        res = rgb_as_hex(self.name[1:7])
        if self.opacity:
            res += hex(round(float(self.opacity) * 255))[2:]
        return res

    def url(self, req):
        return svg.data_url(svg.icon(self.name, opacity=str(self.opacity)))


#: a list of all available icons:
ICONS = [Icon('%s%s' % (s, c)) for s in SHAPES for c in COLORS]

#: a dictionary mapping icon names to icon objects:
ICON_MAP = {icon.name: icon for icon in ICONS}

#: a list of icons ordered by preference:
ORDERED_ICONS = [ICON_MAP[s + c] for s, c in itertools.chain(
                 itertools.product(SHAPES, PREFERED_COLORS),
                 itertools.product(SHAPES, SECONDARY_COLORS))]


@implementer(IMapMarker)
class MapMarker(object):

    """The default map marker is an orange circle."""

    def get_icon(self, ctx, req):
        return DEFAULT_ICON

    def __call__(self, ctx, req):
        icon = self.get_icon(ctx, req) or DEFAULT_ICON
        return req.registry.getUtility(IIcon, icon).url(req)
